# -*- coding: utf-8 -*-

"""Tests for the BEL parser."""

import logging
import re
import unittest

from pybel import BELGraph
from pybel.constants import (
    ABUNDANCE,
    ACTIVITY,
    BIOPROCESS,
    CELL_SECRETION,
    CELL_SURFACE_EXPRESSION,
    COMPLEX,
    COMPOSITE,
    CONCEPT,
    DEGRADATION,
    DIRECTLY_INCREASES,
    DIRTY,
    EFFECT,
    FRAGMENT,
    FROM_LOC,
    FUNCTION,
    FUSION,
    FUSION_MISSING,
    FUSION_REFERENCE,
    FUSION_START,
    FUSION_STOP,
    GENE,
    HAS_VARIANT,
    HGVS,
    IDENTIFIER,
    KIND,
    LOCATION,
    MEMBERS,
    MIRNA,
    MODIFIER,
    NAME,
    NAMESPACE,
    PART_OF,
    PARTNER_3P,
    PARTNER_5P,
    PATHOLOGY,
    POPULATION,
    PRODUCTS,
    PROTEIN,
    RANGE_3P,
    RANGE_5P,
    REACTANTS,
    REACTION,
    RELATION,
    RNA,
    SOURCE,
    TARGET,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSLOCATION,
    VARIANTS,
)
from pybel.dsl import (
    Fragment,
    Population,
    abundance,
    bioprocess,
    cell_surface_expression,
    complex_abundance,
    composite_abundance,
    fragment,
    fusion_range,
    gene,
    gene_fusion,
    gmod,
    hgvs,
    mirna,
    named_complex_abundance,
    pathology,
    pmod,
    protein,
    protein_fusion,
    reaction,
    rna,
    rna_fusion,
    secretion,
    translocation,
)
from pybel.dsl.namespaces import hgnc
from pybel.exceptions import MalformedTranslocationWarning
from pybel.language import Entity, activity_mapping
from pybel.parser import BELParser
from pybel.parser.parse_bel import modifier_po_to_dict
from tests.constants import (
    TestTokenParserBase,
    assert_has_edge,
    assert_has_node,
    update_provenance,
)

logger = logging.getLogger(__name__)

TEST_GENE_VARIANT = "c.308G>A"
TEST_PROTEIN_VARIANT = "p.Phe508del"


class TestAbundance(TestTokenParserBase):
    """2.1.1"""

    def setUp(self):
        self.parser.clear()
        self.parser.general_abundance.setParseAction(self.parser.handle_term)

        self.expected_node = abundance(namespace="CHEBI", name="oxygen atom")
        self.expected_canonical_bel = 'a(CHEBI:"oxygen atom")'

    def _test_abundance_helper(self, statement):
        result = self.parser.general_abundance.parseString(statement)
        self.assertEqual(dict(self.expected_node), result.asDict())

        self.assertIn(self.expected_node, self.graph)
        self.assertEqual(self.expected_canonical_bel, self.graph.node_to_bel(self.expected_node))

        self.assertEqual(
            {},
            modifier_po_to_dict(result),
            msg="The modifier dictionary should be empty",
        )

    def test_abundance(self):
        """Test short/long abundance name."""
        self._test_abundance_helper('a(CHEBI:"oxygen atom")')
        self._test_abundance_helper('abundance(CHEBI:"oxygen atom")')

    def _test_abundance_with_location_helper(self, statement):
        result = self.parser.general_abundance.parseString(statement)

        expected_result = {
            FUNCTION: ABUNDANCE,
            CONCEPT: {
                NAMESPACE: "CHEBI",
                NAME: "oxygen atom",
            },
            LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
        }

        self.assertEqual(expected_result, result.asDict())

        self.assertIn(self.expected_node, self.graph)
        self.assertEqual(self.expected_canonical_bel, self.graph.node_to_bel(self.expected_node))

        modifier = modifier_po_to_dict(result)
        expected_modifier = {LOCATION: {NAMESPACE: "GO", NAME: "intracellular"}}
        self.assertEqual(expected_modifier, modifier)

    def test_abundance_with_location(self):
        """Test short/long abundance name and short/long location name."""
        self._test_abundance_with_location_helper('a(CHEBI:"oxygen atom", loc(GO:intracellular))')
        self._test_abundance_with_location_helper('abundance(CHEBI:"oxygen atom", loc(GO:intracellular))')
        self._test_abundance_with_location_helper('a(CHEBI:"oxygen atom", location(GO:intracellular))')
        self._test_abundance_with_location_helper('abundance(CHEBI:"oxygen atom", location(GO:intracellular))')


class TestAbundanceLabeled(TestTokenParserBase):
    """2.1.1"""

    def setUp(self):
        self.parser.clear()
        self.parser.general_abundance.setParseAction(self.parser.handle_term)

        self.expected_node = abundance(namespace="CHEBI", name="oxygen atom", identifier="CHEBI:25805")
        self.expected_canonical_bel = 'a(CHEBI:"CHEBI:25805" ! "oxygen atom")'

    def _test_abundance_helper(self, statement):
        result = self.parser.general_abundance.parseString(statement)
        self.assertEqual(dict(self.expected_node), result.asDict())

        self.assertIn(self.expected_node, self.graph)
        node = list(self.graph)[0]
        self.assertEqual(self.expected_canonical_bel, node.as_bel())

        self.assertEqual(
            {},
            modifier_po_to_dict(result),
            msg="The modifier dictionary should be empty",
        )

    def test_abundance(self):
        """Test short/long abundance name."""
        for s in (
            'a(CHEBI:"CHEBI:25805"!"oxygen atom")',
            'abundance(CHEBI:"CHEBI:25805"!"oxygen atom")',
            'abundance(CHEBI:"CHEBI:25805" ! "oxygen atom")',
        ):
            with self.subTest(s=s):
                self._test_abundance_helper(s)

    def _test_abundance_with_location_helper(self, statement):
        result = self.parser.general_abundance.parseString(statement)

        expected_result = {
            FUNCTION: ABUNDANCE,
            CONCEPT: {
                NAMESPACE: "CHEBI",
                NAME: "oxygen atom",
                IDENTIFIER: "CHEBI:25805",
            },
            LOCATION: {
                NAMESPACE: "GO",
                NAME: "intracellular",
            },
        }

        self.assertEqual(expected_result, result.asDict())

        self.assertIn(self.expected_node, self.graph)
        self.assertEqual(self.expected_canonical_bel, self.graph.node_to_bel(self.expected_node))

        modifier = modifier_po_to_dict(result)
        expected_modifier = {
            LOCATION: {
                NAMESPACE: "GO",
                NAME: "intracellular",
            }
        }
        self.assertEqual(expected_modifier, modifier)

    def test_abundance_with_location(self):
        """Test short/long abundance name and short/long location name."""
        self._test_abundance_with_location_helper('a(CHEBI:"CHEBI:25805"!"oxygen atom", loc(GO:intracellular))')
        self._test_abundance_with_location_helper('abundance(CHEBI:"CHEBI:25805"!"oxygen atom", loc(GO:intracellular))')
        self._test_abundance_with_location_helper('a(CHEBI:"CHEBI:25805"!"oxygen atom", location(GO:intracellular))')
        self._test_abundance_with_location_helper(
            'abundance(CHEBI:"CHEBI:25805"!"oxygen atom", location(GO:intracellular))'
        )


class TestGene(TestTokenParserBase):
    """2.1.4 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XgeneA"""

    def setUp(self):
        self.parser.clear()
        self.parser.gene.setParseAction(self.parser.handle_term)

    def test_gene_simple(self):
        """Test parsing a simple gene."""
        statement = "g(HGNC:AKT1)"

        result = self.parser.gene.parseString(statement)
        expected_list = [GENE, ["HGNC", "AKT1"]]
        self.assertEqual(expected_list, result.asList())

        expected_dict = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene("HGNC", "AKT1")
        self.assert_has_node(expected_node)
        self.assertEqual("g(HGNC:AKT1)", self.graph.node_to_bel(expected_node))
        self.assertEqual(1, len(self.graph))

    def test_gene_with_location(self):
        """Test parsing a gene with a location."""
        statement = "g(HGNC:AKT1, loc(GO:intracellular))"

        result = self.parser.gene.parseString(statement)

        expected_dict = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            LOCATION: {
                NAMESPACE: "GO",
                NAME: "intracellular",
            },
        }

        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene("HGNC", "AKT1")
        self.assert_has_node(expected_node)
        self.assertEqual("g(HGNC:AKT1)", self.graph.node_to_bel(expected_node))

    def test_gene_with_hgvs(self):
        """Test parsing a gene with a variant."""
        statement = "g(HGNC:AKT1, var(p.Phe508del))"
        result = self.parser.gene.parseString(statement)

        expected_result = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            VARIANTS: [hgvs(TEST_PROTEIN_VARIANT)],
        }
        self.assertEqual(expected_result, result.asDict())

        expected_node = gene("HGNC", "AKT1", variants=hgvs(TEST_PROTEIN_VARIANT))
        self.assert_has_node(expected_node)
        self.assertEqual('g(HGNC:AKT1, var("p.Phe508del"))', self.graph.node_to_bel(expected_node))

        parent = gene("HGNC", "AKT1")
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_gene_with_gmod(self):
        """Test parsing a gene with a gene modification."""
        statement = "geneAbundance(HGNC:AKT1,gmod(M))"
        result = self.parser.gene.parseString(statement)

        expected_result = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            VARIANTS: [gmod("Me")],
        }
        self.assertEqual(expected_result, result.asDict())

        expected_node = gene("HGNC", "AKT1", variants=gmod("Me"))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(HGNC:AKT1, gmod(go:0006306 ! "DNA methylation"))',
            self.graph.node_to_bel(expected_node),
        )

        parent = gene("HGNC", "AKT1")
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_gene_with_substitution(self):
        """Test BEL 1.0 gene substitution"""
        statement = "g(HGNC:AKT1,sub(G,308,A))"
        result = self.parser.gene.parseString(statement)

        expected_result = gene(
            name="AKT1",
            namespace="HGNC",
            variants=[hgvs(TEST_GENE_VARIANT)],
        )
        self.assertEqual(dict(expected_result), result.asDict())

        expected_node = gene("HGNC", "AKT1", variants=hgvs("c.308G>A"))
        self.assert_has_node(expected_node)
        self.assertEqual('g(HGNC:AKT1, var("c.308G>A"))', self.graph.node_to_bel(expected_node))

        parent = gene("HGNC", "AKT1")
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_gene_with_substitution_and_location(self):
        """Test BEL 1.0 gene substitution with location tag"""
        statement = "g(HGNC:AKT1,sub(G,308,A),loc(GO:intracellular))"
        result = self.parser.gene.parseString(statement)

        expected_result = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            VARIANTS: [
                {
                    KIND: HGVS,
                    HGVS: TEST_GENE_VARIANT,
                },
            ],
            LOCATION: {
                NAMESPACE: "GO",
                NAME: "intracellular",
            },
        }
        self.assertEqual(expected_result, result.asDict())

        expected_node = gene("HGNC", "AKT1", variants=hgvs("c.308G>A"))
        self.assert_has_node(expected_node)
        self.assertEqual('g(HGNC:AKT1, var("c.308G>A"))', self.graph.node_to_bel(expected_node))

        parent = gene("HGNC", "AKT1")
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_multiple_variants(self):
        """Test multiple variants"""
        statement = "g(HGNC:AKT1, var(p.Phe508del), sub(G,308,A), var(c.1521_1523delCTT))"
        result = self.parser.gene.parseString(statement)

        expected_result = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            VARIANTS: [
                hgvs(TEST_PROTEIN_VARIANT),
                hgvs(TEST_GENE_VARIANT),
                hgvs("c.1521_1523delCTT"),
            ],
        }
        self.assertEqual(expected_result, result.asDict())

        expected_node = gene(
            "HGNC",
            "AKT1",
            variants=[
                hgvs("c.1521_1523delCTT"),
                hgvs(TEST_GENE_VARIANT),
                hgvs(TEST_PROTEIN_VARIANT),
            ],
        )
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(HGNC:AKT1, var("c.1521_1523delCTT"), var("c.308G>A"), var("p.Phe508del"))',
            self.graph.node_to_bel(expected_node),
        )

        parent = gene("HGNC", "AKT1")
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def _help_test_gene_fusion_1(self, statement):
        result = self.parser.gene.parseString(statement)
        expected_dict = {
            FUNCTION: GENE,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {
                    FUSION_REFERENCE: "c",
                    FUSION_START: 1,
                    FUSION_STOP: 79,
                },
                RANGE_3P: {
                    FUSION_REFERENCE: "c",
                    FUSION_START: 312,
                    FUSION_STOP: 5034,
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        en = gene_fusion(
            partner_5p=gene("HGNC", "TMPRSS2"),
            range_5p=fusion_range("c", 1, 79),
            partner_3p=gene("HGNC", "ERG"),
            range_3p=fusion_range("c", 312, 5034),
        )
        self.assert_has_node(en)

        self.assertEqual(
            'g(fus(HGNC:TMPRSS2, "c.1_79", HGNC:ERG, "c.312_5034"))',
            self.graph.node_to_bel(en),
        )

    def test_gene_fusion_1(self):
        # no quotes
        self._help_test_gene_fusion_1("g(fus(HGNC:TMPRSS2, c.1_79, HGNC:ERG, c.312_5034))")
        # quotes
        self._help_test_gene_fusion_1('g(fus(HGNC:TMPRSS2, "c.1_79", HGNC:ERG, "c.312_5034"))')

    def _help_test_gene_fusion_2(self, statement):
        result = self.parser.gene.parseString(statement)
        expected_dict = {
            FUNCTION: GENE,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {FUSION_REFERENCE: "c", FUSION_START: 1, FUSION_STOP: "?"},
                RANGE_3P: {FUSION_REFERENCE: "c", FUSION_START: 312, FUSION_STOP: 5034},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene_fusion(
            gene("HGNC", "TMPRSS2"),
            gene("HGNC", "ERG"),
            fusion_range("c", 1, "?"),
            fusion_range("c", 312, 5034),
        )
        self.assert_has_node(expected_node)

        canonical_bel = self.graph.node_to_bel(expected_node)
        self.assertEqual('g(fus(HGNC:TMPRSS2, "c.1_?", HGNC:ERG, "c.312_5034"))', canonical_bel)

    def test_gene_fusion_2(self):
        # no quotes
        self._help_test_gene_fusion_2("g(fus(HGNC:TMPRSS2, c.1_?, HGNC:ERG, c.312_5034))")
        # correct
        self._help_test_gene_fusion_2('g(fus(HGNC:TMPRSS2, "c.1_?", HGNC:ERG, "c.312_5034"))')

    def _help_test_gene_fusion_3(self, statement):
        result = self.parser.gene.parseString(statement)
        expected_dict = {
            FUNCTION: GENE,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {FUSION_MISSING: "?"},
                RANGE_3P: {FUSION_REFERENCE: "c", FUSION_START: 312, FUSION_STOP: 5034},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene_fusion(
            gene("HGNC", "TMPRSS2"),
            gene("HGNC", "ERG"),
            range_3p=fusion_range("c", 312, 5034),
        )
        self.assert_has_node(expected_node)

        self.assertEqual(
            'g(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "c.312_5034"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_gene_fusion_3(self):
        # no quotes
        self._help_test_gene_fusion_3("g(fus(HGNC:TMPRSS2, ?, HGNC:ERG, c.312_5034))")
        # correct
        self._help_test_gene_fusion_3('g(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "c.312_5034"))')

    def _help_test_gene_fusion_legacy_1(self, statement):
        result = self.parser.gene.parseString(statement)

        expected_dict = {
            FUNCTION: GENE,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "BCR"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "JAK2"}},
                RANGE_5P: {FUSION_REFERENCE: "c", FUSION_START: "?", FUSION_STOP: 1875},
                RANGE_3P: {FUSION_REFERENCE: "c", FUSION_START: 2626, FUSION_STOP: "?"},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene_fusion(
            gene("HGNC", "BCR"),
            gene("HGNC", "JAK2"),
            fusion_range("c", "?", 1875),
            fusion_range("c", 2626, "?"),
        )
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(fus(HGNC:BCR, "c.?_1875", HGNC:JAK2, "c.2626_?"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_gene_fusion_legacy_1(self):
        # legacy
        self._help_test_gene_fusion_legacy_1("g(HGNC:BCR, fus(HGNC:JAK2, 1875, 2626))")
        # no quotes
        self._help_test_gene_fusion_legacy_1("g(fus(HGNC:BCR, c.?_1875, HGNC:JAK2, c.2626_?))")
        # correct
        self._help_test_gene_fusion_legacy_1('g(fus(HGNC:BCR, "c.?_1875", HGNC:JAK2, "c.2626_?"))')

    def _help_test_gene_fusion_legacy_2(self, statement):
        result = self.parser.gene.parseString(statement)

        expected_dict = {
            FUNCTION: GENE,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "CHCHD4"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "AIFM1"}},
                RANGE_5P: {FUSION_MISSING: "?"},
                RANGE_3P: {FUSION_MISSING: "?"},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = gene_fusion(gene("HGNC", "CHCHD4"), gene("HGNC", "AIFM1"))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(fus(HGNC:CHCHD4, "?", HGNC:AIFM1, "?"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_gene_fusion_legacy_2(self):
        # legacy
        self._help_test_gene_fusion_legacy_2("g(HGNC:CHCHD4, fusion(HGNC:AIFM1))")
        # no quotes
        self._help_test_gene_fusion_legacy_2("g(fus(HGNC:CHCHD4, ?, HGNC:AIFM1, ?))")
        # correct
        self._help_test_gene_fusion_legacy_2('g(fus(HGNC:CHCHD4, "?", HGNC:AIFM1, "?"))')

    def test_gene_variant_snp(self):
        """2.2.2 SNP"""
        statement = "g(SNP:rs113993960, var(c.1521_1523delCTT))"
        result = self.parser.gene.parseString(statement)

        expected_result = [GENE, ["SNP", "rs113993960"], [HGVS, "c.1521_1523delCTT"]]
        self.assertEqual(expected_result, result.asList())

        expected_node = gene("SNP", "rs113993960", variants=hgvs("c.1521_1523delCTT"))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(SNP:rs113993960, var("c.1521_1523delCTT"))',
            self.graph.node_to_bel(expected_node),
        )

        gene_node = expected_node.get_parent()
        self.assert_has_node(gene_node)
        self.assert_has_edge(gene_node, expected_node, relation=HAS_VARIANT)

    def test_gene_variant_chromosome(self):
        """2.2.2 chromosome"""
        statement = 'g(REF:"NC_000007.13", var(g.117199646_117199648delCTT))'
        result = self.parser.gene.parseString(statement)

        expected_result = [
            GENE,
            ["REF", "NC_000007.13"],
            [HGVS, "g.117199646_117199648delCTT"],
        ]
        self.assertEqual(expected_result, result.asList())

        gene_node = gene("REF", "NC_000007.13", variants=hgvs("g.117199646_117199648delCTT"))
        self.assert_has_node(gene_node)

        parent = gene_node.get_parent()
        self.assert_has_node(parent)
        self.assert_has_edge(parent, gene_node, relation=HAS_VARIANT)

    def test_gene_variant_deletion(self):
        """2.2.2 gene-coding DNA reference sequence"""
        statement = "g(HGNC:CFTR, var(c.1521_1523delCTT))"
        result = self.parser.gene.parseString(statement)

        expected_result = {
            FUNCTION: GENE,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "CFTR",
            },
            VARIANTS: [
                {
                    KIND: HGVS,
                    HGVS: "c.1521_1523delCTT",
                },
            ],
        }
        self.assertEqual(expected_result, result.asDict())

        expected_node = gene("HGNC", "CFTR", variants=hgvs("c.1521_1523delCTT"))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'g(HGNC:CFTR, var("c.1521_1523delCTT"))',
            self.graph.node_to_bel(expected_node),
        )

        gene_node = expected_node.get_parent()
        self.assert_has_node(gene_node)
        self.assert_has_edge(gene_node, expected_node, relation=HAS_VARIANT)


class TestMicroRna(TestTokenParserBase):
    """2.1.5 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XmicroRNAA"""

    def setUp(self):
        self.parser.clear()
        self.parser.mirna.setParseAction(self.parser.handle_term)

    def _test_no_variant_helper(self, statement):
        result = self.parser.mirna.parseString(statement)
        expected_result = [MIRNA, ["HGNC", "MIR21"]]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            FUNCTION: MIRNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "MIR21",
            },
        }

        self.assertEqual(expected_dict, result.asDict())

        node = mirna("HGNC", "MIR21")
        self.assert_has_node(node)

    def test_short(self):
        self._test_no_variant_helper("m(HGNC:MIR21)")
        self._test_no_variant_helper("microRNAAbundance(HGNC:MIR21)")

    def test_mirna_location(self):
        statement = "m(HGNC:MIR21,loc(GO:intracellular))"
        result = self.parser.mirna.parseString(statement)

        expected_dict = {
            FUNCTION: MIRNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "MIR21",
            },
            LOCATION: {
                NAMESPACE: "GO",
                NAME: "intracellular",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = mirna("HGNC", "MIR21")
        self.assert_has_node(expected_node)

    def test_mirna_variant(self):
        statement = "m(HGNC:MIR21,var(p.Phe508del))"
        result = self.parser.mirna.parseString(statement)

        expected_dict = {
            FUNCTION: MIRNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "MIR21",
            },
            VARIANTS: [
                hgvs(TEST_PROTEIN_VARIANT),
            ],
        }
        self.assertEqual(expected_dict, result.asDict())

        node = mirna("HGNC", "MIR21", variants=hgvs(TEST_PROTEIN_VARIANT))
        self.assert_has_node(node)
        self.assertEqual('m(HGNC:MIR21, var("p.Phe508del"))', self.graph.node_to_bel(node))

        self.assertEqual(2, self.parser.graph.number_of_nodes())

        expected_parent = node.get_parent()
        self.assert_has_node(expected_parent)
        self.assert_has_edge(expected_parent, node, relation=HAS_VARIANT)

    def test_mirna_variant_location(self):
        statement = "m(HGNC:MIR21,var(p.Phe508del),loc(GO:intracellular))"
        result = self.parser.mirna.parseString(statement)

        expected_dict = {
            FUNCTION: MIRNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "MIR21",
            },
            VARIANTS: [
                {
                    KIND: HGVS,
                    HGVS: "p.Phe508del",
                },
            ],
            LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
        }
        self.assertEqual(expected_dict, result.asDict())

        node = mirna("HGNC", "MIR21", variants=hgvs(TEST_PROTEIN_VARIANT))
        self.assert_has_node(node)
        self.assertEqual('m(HGNC:MIR21, var("p.Phe508del"))', self.graph.node_to_bel(node))

        self.assertEqual(2, self.parser.graph.number_of_nodes())

        expected_parent = node.get_parent()
        self.assert_has_node(expected_parent)
        self.assert_has_edge(expected_parent, node, relation=HAS_VARIANT)


class TestProtein(TestTokenParserBase):
    """2.1.6 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XproteinA"""

    def setUp(self):
        self.parser.clear()
        self.parser.protein.setParseAction(self.parser.handle_term)

    def _test_reference_helper(self, statement):
        result = self.parser.protein.parseString(statement)
        expected_result = [PROTEIN, ["HGNC", "AKT1"]]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        node = protein("HGNC", "AKT1")
        self.assert_has_node(node)
        self.assertEqual("p(HGNC:AKT1)", self.graph.node_to_bel(node))

    def test_reference(self):
        self._test_reference_helper("p(HGNC:AKT1)")
        self._test_reference_helper("proteinAbundance(HGNC:AKT1)")

    def test_protein_with_location(self):
        statement = "p(HGNC:AKT1, loc(GO:intracellular))"

        result = self.parser.protein.parseString(statement)

        expected_dict = {
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
        }
        self.assertEqual(expected_dict, result.asDict())

        node = protein("HGNC", "AKT1")
        self.assert_has_node(node)
        self.assertEqual("p(HGNC:AKT1)", self.graph.node_to_bel(node))

    def test_multiple_variants(self):
        statement = "p(HGNC:AKT1,sub(A,127,Y),pmod(Ph, Ser),loc(GO:intracellular))"

        result = self.parser.protein.parseString(statement)

        expected_dict = {
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            LOCATION: {NAMESPACE: "GO", NAME: "intracellular"},
            VARIANTS: [hgvs("p.Ala127Tyr"), pmod(name="Ph", code="Ser")],
        }
        self.assertEqual(expected_dict, result.asDict())

        parent = protein("HGNC", "AKT1")
        node = parent.with_variants([hgvs("p.Ala127Tyr"), pmod("Ph", code="Ser")])
        self.assert_has_node(node)
        self.assertEqual(
            'p(HGNC:AKT1, pmod(go:0006468 ! "protein phosphorylation", Ser), var("p.Ala127Tyr"))',
            self.graph.node_to_bel(node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, node, relation=HAS_VARIANT)

    def _help_test_protein_fusion_1(self, statement):
        result = self.parser.protein.parseString(statement)
        expected_dict = {
            FUNCTION: PROTEIN,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {
                    FUSION_REFERENCE: "p",
                    FUSION_START: 1,
                    FUSION_STOP: 79,
                },
                RANGE_3P: {
                    FUSION_REFERENCE: "p",
                    FUSION_START: 312,
                    FUSION_STOP: 5034,
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = protein_fusion(
            protein("HGNC", "TMPRSS2"),
            protein("HGNC", "ERG"),
            fusion_range("p", 1, 79),
            fusion_range("p", 312, 5034),
        )
        self.assert_has_node(expected_node)
        self.assertEqual(
            'p(fus(HGNC:TMPRSS2, "p.1_79", HGNC:ERG, "p.312_5034"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_protein_fusion_1(self):
        # no quotes
        self._help_test_protein_fusion_1("p(fus(HGNC:TMPRSS2, p.1_79, HGNC:ERG, p.312_5034))")
        # quotes
        self._help_test_protein_fusion_1('p(fus(HGNC:TMPRSS2, "p.1_79", HGNC:ERG, "p.312_5034"))')

    def _help_test_protein_fusion_legacy_1(self, statement):
        result = self.parser.protein.parseString(statement)

        expected_dict = {
            FUNCTION: PROTEIN,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "BCR"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "JAK2"}},
                RANGE_5P: {
                    FUSION_REFERENCE: "p",
                    FUSION_START: "?",
                    FUSION_STOP: 1875,
                },
                RANGE_3P: {
                    FUSION_REFERENCE: "p",
                    FUSION_START: 2626,
                    FUSION_STOP: "?",
                },
            },
        }

        self.assertEqual(expected_dict, result.asDict())

        expected_node = protein_fusion(
            protein("HGNC", "BCR"),
            protein("HGNC", "JAK2"),
            fusion_range("p", "?", 1875),
            fusion_range("p", 2626, "?"),
        )
        self.assert_has_node(expected_node)
        canonical_bel = self.graph.node_to_bel(expected_node)
        self.assertEqual('p(fus(HGNC:BCR, "p.?_1875", HGNC:JAK2, "p.2626_?"))', canonical_bel)

    def test_protein_fusion_legacy_1(self):
        # legacy (BEL 1.0)
        self._help_test_protein_fusion_legacy_1("p(HGNC:BCR, fus(HGNC:JAK2, 1875, 2626))")
        # missing quotes
        self._help_test_protein_fusion_legacy_1("p(fus(HGNC:BCR, p.?_1875, HGNC:JAK2, p.2626_?))")
        # correct
        self._help_test_protein_fusion_legacy_1('p(fus(HGNC:BCR, "p.?_1875", HGNC:JAK2, "p.2626_?"))')

    def _help_test_protein_legacy_fusion_2(self, statement):
        result = self.parser.protein.parseString(statement)

        expected_dict = {
            FUNCTION: PROTEIN,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "CHCHD4"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "AIFM1"}},
                RANGE_5P: {FUSION_MISSING: "?"},
                RANGE_3P: {FUSION_MISSING: "?"},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = protein_fusion(
            protein("HGNC", "CHCHD4"),
            protein("HGNC", "AIFM1"),
        )
        self.assert_has_node(expected_node)

        canonical_bel = self.graph.node_to_bel(expected_node)
        self.assertEqual('p(fus(HGNC:CHCHD4, "?", HGNC:AIFM1, "?"))', canonical_bel)

    def test_protein_fusion_legacy_2(self):
        # legacy (BEL 1.0)
        self._help_test_protein_legacy_fusion_2("proteinAbundance(HGNC:CHCHD4, fusion(HGNC:AIFM1))")
        # legacy shorthand (BEL 1.0)
        self._help_test_protein_legacy_fusion_2("p(HGNC:CHCHD4, fus(HGNC:AIFM1))")
        # missing quotes
        self._help_test_protein_legacy_fusion_2("p(fus(HGNC:CHCHD4, ?, HGNC:AIFM1, ?))")
        # correct
        self._help_test_protein_legacy_fusion_2('p(fus(HGNC:CHCHD4, "?", HGNC:AIFM1, "?"))')

    def _help_test_protein_trunc_1(self, statement):
        result = self.parser.protein.parseString(statement)

        expected_node = protein("HGNC", "AKT1", variants=hgvs("p.40*"))
        self.assert_has_node(expected_node)

        canonical_bel = self.graph.node_to_bel(expected_node)
        self.assertEqual('p(HGNC:AKT1, var("p.40*"))', canonical_bel)

        protein_node = expected_node.get_parent()
        self.assert_has_node(protein_node)
        self.assert_has_edge(protein_node, expected_node, relation=HAS_VARIANT)

    def test_protein_trunc_1(self):
        # legacy
        self._help_test_protein_trunc_1("p(HGNC:AKT1, trunc(40))")
        # missing quotes
        self._help_test_protein_trunc_1("p(HGNC:AKT1, var(p.40*))")
        # correct
        self._help_test_protein_trunc_1('p(HGNC:AKT1, var("p.40*"))')

    def test_protein_trunc_2(self):
        statement = "p(HGNC:AKT1, var(p.Cys40*))"
        result = self.parser.protein.parseString(statement)

        expected_result = [PROTEIN, ["HGNC", "AKT1"], [HGVS, "p.Cys40*"]]
        self.assertEqual(expected_result, result.asList())

        parent = protein("HGNC", "AKT1")
        expected_node = parent.with_variants(hgvs("p.Cys40*"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:AKT1, var("p.Cys40*"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_trunc_3(self):
        statement = "p(HGNC:AKT1, var(p.Arg1851*))"
        result = self.parser.protein.parseString(statement)

        expected_result = [PROTEIN, ["HGNC", "AKT1"], [HGVS, "p.Arg1851*"]]
        self.assertEqual(expected_result, result.asList())

        parent = protein("HGNC", "AKT1")
        expected_node = parent.with_variants(hgvs("p.Arg1851*"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:AKT1, var("p.Arg1851*"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_pmod_1(self):
        """2.2.1 Test default BEL namespace and 1-letter amino acid code:"""
        statement = "p(HGNC:AKT1, pmod(Ph, S, 473))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "AKT1")
        expected_node = parent.with_variants(pmod("Ph", code="Ser", position=473))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'p(HGNC:AKT1, pmod(go:0006468 ! "protein phosphorylation", Ser, 473))',
            self.graph.node_to_bel(expected_node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_pmod_2(self):
        """2.2.1 Test default BEL namespace and 3-letter amino acid code:"""
        statement = "p(HGNC:AKT1, pmod(Ph, Ser, 473))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "AKT1")
        expected_node = parent.with_variants(pmod("Ph", code="Ser", position=473))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'p(HGNC:AKT1, pmod(go:0006468 ! "protein phosphorylation", Ser, 473))',
            self.graph.node_to_bel(expected_node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_pmod_3(self):
        """2.2.1 Test PSI-MOD namespace and 3-letter amino acid code:"""
        statement = "p(HGNC:AKT1, pmod(MOD:PhosRes,Ser,473))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "AKT1")
        expected_node = parent.with_variants(pmod(namespace="MOD", name="PhosRes", code="Ser", position=473))
        self.assert_has_node(expected_node)
        self.assertEqual(
            "p(HGNC:AKT1, pmod(MOD:PhosRes, Ser, 473))",
            self.graph.node_to_bel(expected_node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_pmod_4(self):
        """2.2.1 Test HRAS palmitoylated at an unspecified residue. Default BEL namespace"""
        statement = "p(HGNC:HRAS,pmod(Palm))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "HRAS")
        expected_node = parent.with_variants(pmod("Palm"))
        self.assert_has_node(expected_node)
        self.assertEqual(
            'p(HGNC:HRAS, pmod(go:0018345 ! "protein palmitoylation"))',
            self.graph.node_to_bel(expected_node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_variant_reference(self):
        """2.2.2 Test reference allele"""
        statement = "p(HGNC:CFTR, var(=))"
        result = self.parser.protein.parseString(statement)
        expected_result = [PROTEIN, ["HGNC", "CFTR"], [HGVS, "="]]
        self.assertEqual(expected_result, result.asList())

        expected_node = protein("HGNC", "CFTR", variants=hgvs("="))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:CFTR, var("="))', self.graph.node_to_bel(expected_node))

        protein_node = expected_node.get_parent()
        self.assert_has_node(protein_node)
        self.assert_has_edge(protein_node, expected_node, relation=HAS_VARIANT)

    def test_protein_variant_unspecified(self):
        """2.2.2 Test unspecified variant"""
        statement = "p(HGNC:CFTR, var(?))"
        result = self.parser.protein.parseString(statement)

        expected_result = [PROTEIN, ["HGNC", "CFTR"], [HGVS, "?"]]
        self.assertEqual(expected_result, result.asList())

        expected_node = protein("HGNC", "CFTR", variants=hgvs("?"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:CFTR, var("?"))', self.graph.node_to_bel(expected_node))

        parent = expected_node.get_parent()
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_variant_substitution(self):
        """2.2.2 Test substitution"""
        statement = "p(HGNC:CFTR, var(p.Gly576Ala))"
        result = self.parser.protein.parseString(statement)
        expected_result = [PROTEIN, ["HGNC", "CFTR"], [HGVS, "p.Gly576Ala"]]
        self.assertEqual(expected_result, result.asList())

        expected_node = protein("HGNC", "CFTR", variants=hgvs("p.Gly576Ala"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:CFTR, var("p.Gly576Ala"))', self.graph.node_to_bel(expected_node))

        parent = expected_node.get_parent()
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_variant_deletion(self):
        """2.2.2 deletion"""
        statement = "p(HGNC:CFTR, var(p.Phe508del))"
        result = self.parser.protein.parseString(statement)

        expected_result = [PROTEIN, ["HGNC", "CFTR"], [HGVS, TEST_PROTEIN_VARIANT]]
        self.assertEqual(expected_result, result.asList())

        expected_node = protein("HGNC", "CFTR", variants=hgvs("p.Phe508del"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:CFTR, var("p.Phe508del"))', self.graph.node_to_bel(expected_node))

        protein_node = expected_node.get_parent()
        self.assert_has_node(protein_node)
        self.assert_has_edge(protein_node, expected_node, relation=HAS_VARIANT)

    def test_protein_fragment_known(self):
        """2.2.3 fragment with known start/stop"""
        statement = "p(HGNC:YFG, frag(5_20))"
        self.parser.protein.parseString(statement)

        parent = protein("HGNC", "YFG")
        expected_node = parent.with_variants(fragment(5, 20))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:YFG, frag("5_20"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_fragment_unbounded(self):
        """2.2.3 amino-terminal fragment of unknown length"""
        statement = "p(HGNC:YFG, frag(1_?))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "YFG")
        expected_node = parent.with_variants(fragment(1, "?"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:YFG, frag("1_?"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_fragment_unboundTerminal(self):
        """2.2.3 carboxyl-terminal fragment of unknown length"""
        statement = "p(HGNC:YFG, frag(?_*))"
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "YFG")
        expected_node = parent.with_variants(fragment("?", "*"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:YFG, frag("?_*"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_fragment_unknown(self):
        """2.2.3 fragment with unknown start/stop"""
        statement = "p(HGNC:YFG, frag(?))"
        result = self.parser.protein.parseString(statement)

        expected_result = [PROTEIN, ["HGNC", "YFG"], [FRAGMENT, "?"]]
        self.assertEqual(expected_result, result.asList())

        parent = protein("HGNC", "YFG")
        expected_node = parent.with_variants(fragment())
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:YFG, frag("?"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_protein_fragment_descriptor(self):
        """2.2.3 fragment with unknown start/stop and a descriptor"""
        statement = 'p(HGNC:YFG, frag(?, "55kD"))'
        result = self.parser.protein.parseString(statement)

        parent = protein("HGNC", "YFG")
        expected_node = parent.with_variants(fragment("?", description="55kD"))
        self.assert_has_node(expected_node)
        self.assertEqual('p(HGNC:YFG, frag("?", "55kD"))', self.graph.node_to_bel(expected_node))

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def test_ensure_no_dup_edges(self):
        """Ensure node and edges aren't added twice, even if from different statements and has origin completion"""
        s1 = "p(HGNC:AKT1)"
        s2 = "deg(p(HGNC:AKT1))"
        node = protein("HGNC", "AKT1")

        self.parser.bel_term.parseString(s1)
        self.assert_has_node(node)
        self.assertEqual(1, self.parser.graph.number_of_nodes())
        self.assertEqual(0, self.parser.graph.number_of_edges())

        self.parser.bel_term.parseString(s2)
        self.assert_has_node(node)
        self.assertEqual(1, self.parser.graph.number_of_nodes())
        self.assertEqual(0, self.parser.graph.number_of_edges())


class TestRna(TestTokenParserBase):
    """2.1.7 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XrnaA"""

    def setUp(self):
        self.parser.clear()
        self.parser.rna.setParseAction(self.parser.handle_term)

    def _help_test_reference(self, statement):
        result = self.parser.rna.parseString(statement)
        expected_result = [RNA, ["HGNC", "AKT1"]]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            FUNCTION: RNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = rna("HGNC", "AKT1")
        self.assert_has_node(expected_node)

        self.assertEqual("r(HGNC:AKT1)", self.graph.node_to_bel(expected_node))

    def test_reference(self):
        # short
        self._help_test_reference("r(HGNC:AKT1)")
        # long
        self._help_test_reference("rnaAbundance(HGNC:AKT1)")

    def test_multiple_variants(self):
        """Test multiple variants."""
        statement = "r(HGNC:AKT1, var(p.Phe508del), var(c.1521_1523delCTT))"
        result = self.parser.rna.parseString(statement)

        expected_result = {
            FUNCTION: RNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
            VARIANTS: [
                hgvs(TEST_PROTEIN_VARIANT),
                hgvs("c.1521_1523delCTT"),
            ],
        }
        self.assertEqual(expected_result, result.asDict())

        parent = rna("HGNC", "AKT1")
        expected_node = parent.with_variants([hgvs("c.1521_1523delCTT"), hgvs(TEST_PROTEIN_VARIANT)])
        self.assert_has_node(expected_node)

        self.assertEqual(
            'r(HGNC:AKT1, var("c.1521_1523delCTT"), var("p.Phe508del"))',
            self.graph.node_to_bel(expected_node),
        )

        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)

    def _help_test_rna_fusion_1(self, statement):
        result = self.parser.rna.parseString(statement)

        expected_dict = {
            FUNCTION: RNA,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {
                    FUSION_REFERENCE: "r",
                    FUSION_START: 1,
                    FUSION_STOP: 79,
                },
                RANGE_3P: {
                    FUSION_REFERENCE: "r",
                    FUSION_START: 312,
                    FUSION_STOP: 5034,
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = rna_fusion(
            rna("HGNC", "TMPRSS2"),
            rna("HGNC", "ERG"),
            fusion_range("r", 1, 79),
            fusion_range("r", 312, 5034),
        )
        self.assert_has_node(expected_node)

        self.assertEqual(
            'r(fus(HGNC:TMPRSS2, "r.1_79", HGNC:ERG, "r.312_5034"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_rna_fusion_known_breakpoints(self):
        """Test RNA fusions (2.6.1) with known breakpoints (2.6.1)."""
        # missing quotes
        self._help_test_rna_fusion_1("r(fus(HGNC:TMPRSS2, r.1_79, HGNC:ERG, r.312_5034))")
        # correct (short form)
        self._help_test_rna_fusion_1('r(fus(HGNC:TMPRSS2, "r.1_79", HGNC:ERG, "r.312_5034"))')
        # correct (long form)
        self._help_test_rna_fusion_1('rnaAbundance(fusion(HGNC:TMPRSS2, "r.1_79", HGNC:ERG, "r.312_5034"))')

    def _help_test_rna_fusion_unspecified_breakpoints(self, statement):
        result = self.parser.rna.parseString(statement)

        expected_dict = {
            FUNCTION: RNA,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "TMPRSS2"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "ERG"}},
                RANGE_5P: {
                    FUSION_MISSING: "?",
                },
                RANGE_3P: {
                    FUSION_MISSING: "?",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = rna_fusion(
            rna("HGNC", "TMPRSS2"),
            rna("HGNC", "ERG"),
        )
        self.assert_has_node(expected_node)
        self.assertEqual(
            'r(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_rna_fusion_unspecified_breakpoints(self):
        """Test RNA fusions (2.6.1) with unspecified breakpoints."""
        # legacy
        self._help_test_rna_fusion_unspecified_breakpoints("r(HGNC:TMPRSS2, fusion(HGNC:ERG))")
        # missing quotes
        self._help_test_rna_fusion_unspecified_breakpoints("r(fus(HGNC:TMPRSS2, ?, HGNC:ERG, ?))")
        # correct (short form)
        self._help_test_rna_fusion_unspecified_breakpoints('r(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))')
        # correct (long form)
        self._help_test_rna_fusion_unspecified_breakpoints('rnaAbundance(fusion(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))')

    def _help_test_rna_fusion_legacy_1(self, statement):
        result = self.parser.rna.parseString(statement)

        expected_dict = {
            FUNCTION: RNA,
            FUSION: {
                PARTNER_5P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "BCR"}},
                PARTNER_3P: {CONCEPT: {NAMESPACE: "HGNC", NAME: "JAK2"}},
                RANGE_5P: {
                    FUSION_REFERENCE: "r",
                    FUSION_START: "?",
                    FUSION_STOP: 1875,
                },
                RANGE_3P: {
                    FUSION_REFERENCE: "r",
                    FUSION_START: 2626,
                    FUSION_STOP: "?",
                },
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = rna_fusion(
            rna("HGNC", "BCR"),
            rna("HGNC", "JAK2"),
            fusion_range("r", "?", 1875),
            fusion_range("r", 2626, "?"),
        )
        self.assert_has_node(expected_node)
        self.assertEqual(
            'r(fus(HGNC:BCR, "r.?_1875", HGNC:JAK2, "r.2626_?"))',
            self.graph.node_to_bel(expected_node),
        )

    def test_rna_fusion_legacy_1(self):
        # legacy
        self._help_test_rna_fusion_legacy_1("r(HGNC:BCR, fus(HGNC:JAK2, 1875, 2626))")
        # no quotes
        self._help_test_rna_fusion_legacy_1("r(fus(HGNC:BCR, r.?_1875, HGNC:JAK2, r.2626_?))")
        # correct
        self._help_test_rna_fusion_legacy_1('r(fus(HGNC:BCR, "r.?_1875", HGNC:JAK2, "r.2626_?"))')

    def test_rna_variant_codingReference(self):
        """2.2.2 RNA coding reference sequence"""
        statement = "r(HGNC:CFTR, var(r.1521_1523delcuu))"
        result = self.parser.rna.parseString(statement)

        expected_dict = {
            FUNCTION: RNA,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "CFTR",
            },
            VARIANTS: [hgvs("r.1521_1523delcuu")],
        }
        self.assertEqual(expected_dict, result.asDict())

        parent = rna("HGNC", "CFTR")
        expected_node = parent.with_variants(hgvs("r.1521_1523delcuu"))
        self.assert_has_node(expected_node)

        self.assertEqual(
            'r(HGNC:CFTR, var("r.1521_1523delcuu"))',
            self.graph.node_to_bel(expected_node),
        )
        self.assert_has_node(parent)
        self.assert_has_edge(parent, expected_node, relation=HAS_VARIANT)


class TestComplex(TestTokenParserBase):
    """2.1.2 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XcomplexA"""

    def setUp(self):
        self.parser.clear()
        self.parser.complex_abundances.setParseAction(self.parser.handle_term)

    def test_named_complex_singleton(self):
        statement = "complex(FPLX:AP1)"
        result = self.parser.complex_abundances.parseString(statement)

        expected_dict = {
            FUNCTION: COMPLEX,
            CONCEPT: {
                NAMESPACE: "FPLX",
                NAME: "AP1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = named_complex_abundance("FPLX", "AP1")
        self.assert_has_node(expected_node)

    def test_complex_list_short(self):
        statement = "complex(p(HGNC:FOS), p(HGNC:JUN))"
        result = self.parser.complex_abundances.parseString(statement)

        expected_result = [
            COMPLEX,
            [PROTEIN, ["HGNC", "FOS"]],
            [PROTEIN, ["HGNC", "JUN"]],
        ]
        self.assertEqual(expected_result, result.asList())

        expected_result = {
            FUNCTION: COMPLEX,
            MEMBERS: [
                {
                    FUNCTION: PROTEIN,
                    CONCEPT: {
                        NAMESPACE: "HGNC",
                        NAME: "FOS",
                    },
                },
                {
                    FUNCTION: PROTEIN,
                    CONCEPT: {
                        NAMESPACE: "HGNC",
                        NAME: "JUN",
                    },
                },
            ],
        }
        self.assertEqual(expected_result, result.asDict())

        child_1 = protein("HGNC", "FOS")
        self.assert_has_node(child_1)

        child_2 = protein("HGNC", "JUN")
        self.assert_has_node(child_2)

        expected_node = complex_abundance([child_1, child_2])
        self.assert_has_node(expected_node)

        self.assert_has_edge(child_1, expected_node, relation=PART_OF)
        self.assert_has_edge(child_2, expected_node, relation=PART_OF)

    def test_complex_list_long(self):
        statement = "complexAbundance(proteinAbundance(HGNC:HBP1),geneAbundance(HGNC:NCF1))"
        self.parser.complex_abundances.parseString(statement)


class TestComposite(TestTokenParserBase):
    """Tests the parsing of the composite function

    .. seealso::

            `BEL 2.0 Specification 2.1.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcompositeA>`_
    """

    def setUp(self):
        self.parser.clear()
        self.parser.composite_abundance.setParseAction(self.parser.handle_term)

    def test_213a(self):
        """Evidence: ``IL-6 and IL-23 synergistically induce Th17 differentiation"""
        statement = 'composite(p(HGNC:IL6), complex(GO:"interleukin-23 complex"))'
        result = self.parser.composite_abundance.parseString(statement)

        expected_result = [
            COMPOSITE,
            [PROTEIN, ["HGNC", "IL6"]],
            [COMPLEX, ["GO", "interleukin-23 complex"]],
        ]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            FUNCTION: COMPOSITE,
            MEMBERS: [
                {
                    FUNCTION: PROTEIN,
                    CONCEPT: {
                        NAMESPACE: "HGNC",
                        NAME: "IL6",
                    },
                },
                {
                    FUNCTION: COMPLEX,
                    CONCEPT: {
                        NAMESPACE: "GO",
                        NAME: "interleukin-23 complex",
                    },
                },
            ],
        }
        self.assertEqual(expected_dict, result.asDict())

        il23 = named_complex_abundance("GO", "interleukin-23 complex")
        il6 = protein("HGNC", "IL6")
        expected_node = composite_abundance([il23, il6])
        self.assert_has_node(expected_node)

        self.assertEqual(2, len(expected_node[MEMBERS]))
        self.assertEqual(il23, expected_node[MEMBERS][0])
        self.assertEqual(il6, expected_node[MEMBERS][1])

        self.assertEqual(
            'composite(complex(GO:"interleukin-23 complex"), p(HGNC:IL6))',
            self.graph.node_to_bel(expected_node),
        )

        self.assertEqual(3, self.parser.graph.number_of_nodes())
        self.assert_has_node(expected_node)
        self.assert_has_node(il23)
        self.assert_has_node(il6)
        self.assertEqual(2, self.parser.graph.number_of_edges())


class TestBiologicalProcess(TestTokenParserBase):
    def setUp(self):
        self.parser.clear()
        self.parser.biological_process.setParseAction(self.parser.handle_term)

    def test_231a(self):
        statement = 'bp(GO:"cell cycle arrest")'
        result = self.parser.biological_process.parseString(statement)

        expected_result = [BIOPROCESS, ["GO", "cell cycle arrest"]]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            FUNCTION: BIOPROCESS,
            CONCEPT: {
                NAMESPACE: "GO",
                NAME: "cell cycle arrest",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = bioprocess("GO", "cell cycle arrest")
        self.assert_has_node(expected_node)


class TestPathology(TestTokenParserBase):
    def setUp(self):
        self.parser.clear()
        self.parser.pathology.setParseAction(self.parser.handle_term)

    def test_232a(self):
        statement = "pathology(MESH:adenocarcinoma)"
        result = self.parser.pathology.parseString(statement)

        expected_dict = {
            FUNCTION: PATHOLOGY,
            CONCEPT: {
                NAMESPACE: "MESH",
                NAME: "adenocarcinoma",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = pathology("MESH", "adenocarcinoma")
        self.assert_has_node(expected_node)

        self.assertEqual("path(MESH:adenocarcinoma)", self.graph.node_to_bel(expected_node))


class TestPopulation(TestTokenParserBase):
    def setUp(self):
        self.parser.clear()
        self.parser.population.setParseAction(self.parser.handle_term)

    def test_parse_population(self):
        statement = "pop(UBERON:blood)"
        result = self.parser.population.parseString(statement)

        expected_dict = {
            FUNCTION: POPULATION,
            CONCEPT: {
                NAMESPACE: "UBERON",
                NAME: "blood",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        expected_node = Population("UBERON", "blood")
        self.assert_has_node(expected_node)

        self.assertEqual(
            "pop(UBERON:blood)",
            self.graph.node_to_bel(expected_node),
            msg="Nodes: {}".format(list(self.graph)),
        )


class TestActivity(TestTokenParserBase):
    """Tests for molecular activity terms."""

    def setUp(self):
        """Set up parser for testing the activity language."""
        self.parser.clear()
        self.parser.activity.setParseAction(self.parser.handle_term)

    def test_activity_bare(self):
        statement = "act(p(HGNC:AKT1))"
        result = self.parser.activity.parseString(statement)

        expected_result = [ACTIVITY, PROTEIN, ["HGNC", "AKT1"]]
        self.assertEqual(expected_result, result.asList())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: ACTIVITY,
        }
        self.assertEqual(expected_mod, mod)

    def test_activity_withMolecularActivityDefault(self):
        """Tests activity modifier with molecular activity from default BEL namespace"""
        statement = "act(p(HGNC:AKT1), ma(kin))"
        result = self.parser.activity.parseString(statement)

        expected_dict = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["kin"],
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["kin"],
        }
        self.assertEqual(expected_mod, mod)

    def test_activity_withMolecularActivityDefaultLong(self):
        """Tests activity modifier with molecular activity from custom namespaced"""
        statement = "act(p(HGNC:AKT1), ma(catalyticActivity))"
        result = self.parser.activity.parseString(statement)

        expected_dict = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["cat"],
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["cat"],
        }
        self.assertEqual(expected_mod, mod)

    def test_activity_withMolecularActivityCustom(self):
        """Tests activity modifier with molecular activity from custom namespaced"""
        statement = 'act(p(HGNC:AKT1), ma(GO:"catalytic activity"))'
        result = self.parser.activity.parseString(statement)

        expected_dict = {
            MODIFIER: ACTIVITY,
            EFFECT: {
                NAMESPACE: "GO",
                NAME: "catalytic activity",
            },
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: ACTIVITY,
            EFFECT: {NAMESPACE: "GO", NAME: "catalytic activity"},
        }
        self.assertEqual(expected_mod, mod)

    def test_activity_legacy(self):
        """Test BEL 1.0 style molecular activity annotation"""
        statement = "kin(p(HGNC:AKT1))"
        result = self.parser.activity.parseString(statement)

        expected_dict = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["kin"],
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: ACTIVITY,
            EFFECT: activity_mapping["kin"],
        }
        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "AKT1")
        self.assert_has_node(node)

    def test_kinase_activity_on_named_complex(self):
        statement = "kin(complex(FPLX:C1))"
        self.parser.activity.parseString(statement)

    def test_activity_on_named_complex(self):
        statement = "act(complex(FPLX:C1), ma(kin))"
        self.parser.activity.parseString(statement)

    def test_kinase_activity_on_listed_complex(self):
        statement = "kin(complex(p(HGNC:A), p(HGNC:B)))"
        self.parser.activity.parseString(statement)

    def test_activity_on_listed_complex(self):
        statement = "act(complex(p(HGNC:A), p(HGNC:B)), ma(kin))"
        self.parser.activity.parseString(statement)


class TestTranslocationPermissive(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.graph = BELGraph()
        cls.parser = BELParser(
            cls.graph,
            disallow_unqualified_translocations=False,
            namespace_to_pattern={
                "HGNC": re.compile(r".*"),
                "CHEBI": re.compile(r".*"),
            },
        )

    def setUp(self):
        self.parser.clear()
        self.parser.transformation.setParseAction(self.parser.handle_term)

    def assert_has_node(self, member, **kwargs):
        assert_has_node(self, member, self.parser.graph, **kwargs)

    def assert_has_edge(self, u, v, **kwargs):
        assert_has_edge(self, u, v, self.parser.graph, **kwargs)

    def test_unqualified_translocation_single(self):
        """translocation example"""
        statement = "tloc(p(HGNC:EGFR))"
        result = self.parser.transformation.parseString(statement)

        expected_dict = {
            MODIFIER: TRANSLOCATION,
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "EGFR",
            },
        }

        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: TRANSLOCATION,
        }
        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_unqualified_translocation_relation(self):
        """Test translocation in object.

        3.1.2 http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#XdIncreases
        """
        update_provenance(self.parser.control_parser)

        statement = 'a(CHEBI:"Abeta_42") => tloc(a(CHEBI:"calcium(2+)"))'
        result = self.parser.relation.parseString(statement)

        expected_dict = {
            SOURCE: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "Abeta_42",
                },
            },
            RELATION: DIRECTLY_INCREASES,
            TARGET: {
                FUNCTION: ABUNDANCE,
                CONCEPT: {
                    NAMESPACE: "CHEBI",
                    NAME: "calcium(2+)",
                },
                MODIFIER: TRANSLOCATION,
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        sub = abundance("CHEBI", "Abeta_42")
        self.assert_has_node(sub)

        obj = abundance("CHEBI", "calcium(2+)")
        self.assert_has_node(obj)

        expected_annotations = {
            RELATION: DIRECTLY_INCREASES,
            TARGET_MODIFIER: {
                MODIFIER: TRANSLOCATION,
            },
        }

        self.assert_has_edge(sub, obj, **expected_annotations)


class TestTransformation(TestTokenParserBase):
    def setUp(self):
        self.parser.clear()
        self.parser.transformation.setParseAction(self.parser.handle_term)

    def test_degradation_short(self):
        """Test the short form of degradation works"""
        statement = "deg(p(HGNC:AKT1))"
        result = self.parser.transformation.parseString(statement)

        expected_result = [DEGRADATION, PROTEIN, ["HGNC", "AKT1"]]
        self.assertEqual(expected_result, result.asList())

        expected_dict = {
            MODIFIER: DEGRADATION,
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "AKT1",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: DEGRADATION,
        }
        self.assertEqual(expected_mod, mod)

    def test_degradation_long(self):
        """Test the long form of degradation works"""
        statement = "degradation(p(HGNC:EGFR))"
        result = self.parser.transformation.parseString(statement)

        expected_dict = {
            MODIFIER: DEGRADATION,
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "EGFR",
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: DEGRADATION,
        }
        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_translocation_standard(self):
        """translocation example"""
        statement = 'tloc(p(HGNC:EGFR), fromLoc(GO:"cell surface"), toLoc(GO:endosome))'
        result = self.parser.transformation.parseString(statement)

        expected_dict = {
            MODIFIER: TRANSLOCATION,
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "EGFR",
            },
            EFFECT: {
                FROM_LOC: {
                    NAMESPACE: "GO",
                    NAME: "cell surface",
                },
                TO_LOC: {
                    NAMESPACE: "GO",
                    NAME: "endosome",
                },
            },
        }

        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = translocation(
            from_loc=Entity(namespace="GO", name="cell surface"),
            to_loc=Entity(namespace="GO", name="endosome"),
        )

        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_translocation_bare(self):
        """translocation example"""
        statement = 'tloc(p(HGNC:EGFR), GO:"cell surface", GO:endosome)'
        result = self.parser.transformation.parseString(statement)

        expected_dict = {
            MODIFIER: TRANSLOCATION,
            FUNCTION: PROTEIN,
            CONCEPT: {
                NAMESPACE: "HGNC",
                NAME: "EGFR",
            },
            EFFECT: {
                FROM_LOC: {NAMESPACE: "GO", NAME: "cell surface"},
                TO_LOC: {NAMESPACE: "GO", NAME: "endosome"},
            },
        }
        self.assertEqual(expected_dict, result.asDict())

        mod = modifier_po_to_dict(result)
        expected_mod = {
            MODIFIER: TRANSLOCATION,
            EFFECT: {
                FROM_LOC: {NAMESPACE: "GO", NAME: "cell surface"},
                TO_LOC: {NAMESPACE: "GO", NAME: "endosome"},
            },
        }
        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_unqualified_translocation_strict(self):
        """Fail on an improperly written single argument translocation"""
        statement = 'tloc(pop(EFO:"CD8-Positive T-Lymphocytes"))'
        with self.assertRaises(MalformedTranslocationWarning):
            self.parser.translocation.parseString(statement)

    def test_translocation_secretion(self):
        """cell secretion short form"""
        statement = "sec(p(HGNC:EGFR))"
        result = self.parser.transformation.parseString(statement)

        expected_result = [CELL_SECRETION, PROTEIN, ["HGNC", "EGFR"]]
        self.assertEqual(expected_result, result.asList())

        mod = modifier_po_to_dict(result)
        expected_mod = secretion()
        self.assertEqual(expected_mod, mod)

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_translocation_surface(self):
        """cell surface expression short form"""
        statement = "surf(p(HGNC:EGFR))"
        result = self.parser.transformation.parseString(statement)

        expected_result = [CELL_SURFACE_EXPRESSION, PROTEIN, ["HGNC", "EGFR"]]
        self.assertEqual(expected_result, result.asList())

        expected_mod = cell_surface_expression()
        self.assertEqual(expected_mod, modifier_po_to_dict(result))

        node = protein("HGNC", "EGFR")
        self.assert_has_node(node)

    def test_reaction_1(self):
        statement = 'rxn(reactants(a(CHEBI:superoxide)), products(a(CHEBI:"hydrogen peroxide"), a(CHEBI:oxygen)))'
        result = self.parser.transformation.parseString(statement)

        expected_dict = {
            FUNCTION: REACTION,
            REACTANTS: [
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: "CHEBI",
                        NAME: "superoxide",
                    },
                }
            ],
            PRODUCTS: [
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: "CHEBI",
                        NAME: "hydrogen peroxide",
                    },
                },
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: "CHEBI",
                        NAME: "oxygen",
                    },
                },
            ],
        }
        self.assertEqual(expected_dict, result.asDict())

        superoxide_node = abundance("CHEBI", "superoxide")
        hydrogen_peroxide = abundance("CHEBI", "hydrogen peroxide")
        oxygen_node = abundance("CHEBI", "oxygen")
        expected_node = reaction([superoxide_node], [hydrogen_peroxide, oxygen_node])

        self.assert_has_node(expected_node)

        self.assertEqual(statement, self.graph.node_to_bel(expected_node))

        self.assert_has_node(superoxide_node)
        self.assert_has_edge(expected_node, superoxide_node)

        self.assert_has_node(hydrogen_peroxide)
        self.assert_has_edge(expected_node, hydrogen_peroxide)

        self.assert_has_node(oxygen_node)
        self.assert_has_edge(expected_node, oxygen_node)

    def test_reaction_2(self):
        statement = "rxn(reactants(p(HGNC:APP)), products(p(HGNC:APP, frag(672_713))))"
        self.parser.transformation.parseString(statement)

        app = hgnc(name="APP")
        self.assertIn(app, self.graph)

        amyloid_beta_42 = app.with_variants(Fragment(start=672, stop=713))
        self.assertIn(amyloid_beta_42, self.graph)

        expected_node = reaction(app, amyloid_beta_42)
        self.assertIn(expected_node, self.graph)

    def test_clearance(self):
        """Tests that after adding things, the graph and parser can be cleared properly"""
        s1 = "surf(p(HGNC:EGFR))"
        s2 = 'rxn(reactants(a(CHEBI:superoxide)),products(a(CHEBI:"hydrogen peroxide"), a(CHEBI:"oxygen")))'

        self.parser.transformation.parseString(s1)
        self.parser.transformation.parseString(s2)
        self.assertGreater(self.parser.graph.number_of_nodes(), 0)
        self.assertGreater(self.parser.graph.number_of_edges(), 0)

        self.parser.clear()
        self.assertEqual(0, self.parser.graph.number_of_nodes())
        self.assertEqual(0, self.parser.graph.number_of_edges())
        self.assertEqual(0, len(self.parser.control_parser.annotations))
        self.assertFalse(self.parser.control_parser.citation_is_set)


class TestSemantics(unittest.TestCase):
    def test_lenient_semantic_no_failure(self):
        graph = BELGraph()
        parser = BELParser(graph, allow_naked_names=True)

        update_provenance(parser.control_parser)

        parser.bel_term.addParseAction(parser.handle_term)
        parser.bel_term.parseString("bp(ABASD)")

        node = bioprocess(namespace=DIRTY, name="ABASD")
        self.assertIn(node, graph)



# -*- coding: utf-8 -*-

"""Functions for inducing random sub-graphs."""

import bisect
import logging
import random
from operator import itemgetter
from typing import Any, Iterable, Mapping, Optional, Set, Tuple

from ..utils import remove_isolated_nodes
from ...graph import BELGraph
from ...pipeline import transformation
from ...utils import update_metadata
from ....dsl import BaseEntity

__all__ = [
    "get_graph_with_random_edges",
    "get_random_node",
    "get_random_subgraph",
]

logger = logging.getLogger(__name__)


def _random_edge_iterator(graph: BELGraph, n_edges: int) -> Iterable[Tuple[BaseEntity, BaseEntity, int, Mapping]]:
    """Get a random set of edges from the graph and randomly samples a key from each.

    :param graph: A BEL graph
    :param n_edges: Number of edges to randomly select from the given graph
    """
    edges = list(graph.edges())
    edge_sample = random.sample(edges, n_edges)
    for u, v in edge_sample:
        keys = list(graph[u][v])
        k = random.choice(keys)
        yield u, v, k, graph[u][v][k]


@transformation
def get_graph_with_random_edges(graph: BELGraph, n_edges: int) -> BELGraph:
    """Build a new graph from a seeding of edges.

    :param graph: A BEL graph
    :param n_edges: Number of edges to randomly select from the given graph
    """
    rv = graph.child()
    rv.add_edges_from(_random_edge_iterator(graph, n_edges))
    return rv


#: How many edges should be sampled from a graph that's still reasonable to display
SAMPLE_RANDOM_EDGE_COUNT = 250
#: How many edges should be sampled as "seed" edges
SAMPLE_RANDOM_EDGE_SEED_COUNT = 5


class WeightedRandomGenerator:
    """A weighted random number generator.

    Adapted from: https://eli.thegreenplace.net/2010/01/22/weighted-random-generation-in-python
    """

    def __init__(self, values, weights):
        """Build a weighted random generator.

        :param Any values: A sequence corresponding to the weights
        :param weights: Weights for each. Should all be positive, but not necessarily normalized.
        """
        self.values = values
        self.totals = []
        weight_total = 0

        for weight in weights:
            weight_total += weight
            self.totals.append(weight_total)

    @property
    def total(self):
        """Get the total weight stored."""
        return self.totals[-1]

    def next_index(self) -> int:
        """Get a random index."""
        return bisect.bisect_right(self.totals, random.random() * self.total)

    def next(self) -> Any:
        """Get a random value."""
        return self.values[self.next_index()]


def get_random_node(
    graph,
    node_blacklist: Set[BaseEntity],
    invert_degrees: Optional[bool] = None,
) -> Optional[BaseEntity]:
    """Choose a node from the graph with probabilities based on their degrees.

    :type graph: networkx.Graph
    :param node_blacklist: Nodes to filter out
    :param invert_degrees: Should the degrees be inverted? Defaults to true.
    """
    try:
        nodes, degrees = zip(
            *(
                (node, degree)
                for node, degree in sorted(graph.degree(), key=itemgetter(1))
                if node not in node_blacklist
            ),
        )
    except ValueError:  # something wrong with graph, probably no elements in graph.degree_iter
        return

    if invert_degrees is None or invert_degrees:
        # More likely to choose low degree nodes to explore, so don't make hubs
        degrees = [1 / degree for degree in degrees]

    wrg = WeightedRandomGenerator(nodes, degrees)
    return wrg.next()  # noqa: B305


def _helper(
    result,
    graph,
    number_edges_remaining: int,
    node_blacklist: Set[BaseEntity],
    invert_degrees: Optional[bool] = None,
) -> None:
    """Help build a random graph.

    :type result: networkx.Graph
    :type graph: networkx.Graph
    """
    original_node_count = graph.number_of_nodes()

    logger.debug("adding remaining %d edges", number_edges_remaining)
    for _ in range(number_edges_remaining):

        source, possible_step_nodes, c = None, set(), 0
        while not source or not possible_step_nodes:
            source = get_random_node(result, node_blacklist, invert_degrees=invert_degrees)

            c += 1
            if c >= original_node_count:
                logger.warning("infinite loop happening")
                logger.warning("source: %s", source)
                logger.warning("no grow: %s", node_blacklist)
                return  # Happens when after exhausting the connected components. Try increasing the number seed edges

            if source is None:
                continue  # maybe do something else?

            # Only keep targets in the original graph that aren't in the result graph
            possible_step_nodes = set(graph[source]) - set(result[source])

            if not possible_step_nodes:
                node_blacklist.add(
                    source,
                )  # there aren't any possible nodes to step to, so try growing from somewhere else

        step_node = random.choice(list(possible_step_nodes))

        # it's not really a big deal which, but it might be possible to weight this by the utility of edges later
        key, attr_dict = random.choice(list(graph[source][step_node].items()))

        result.add_edge(source, step_node, key=key, **attr_dict)


@transformation
def get_random_subgraph(
    graph: BELGraph,
    number_edges: Optional[int] = None,
    number_seed_edges: Optional[int] = None,
    seed: Optional[int] = None,
    invert_degrees: Optional[bool] = None,
) -> BELGraph:
    """Generate a random subgraph based on weighted random walks from random seed edges.

    :type graph: pybel.BELGraph graph
    :param number_edges: Maximum number of edges. Defaults to
     :data:`pybel_tools.constants.SAMPLE_RANDOM_EDGE_COUNT` (250).
    :param number_seed_edges: Number of nodes to start with (which likely results in different components
     in large graphs). Defaults to :data:`SAMPLE_RANDOM_EDGE_SEED_COUNT` (5).
    :param seed: A seed for the random state
    :param invert_degrees: Should the degrees be inverted? Defaults to true.
    """
    if number_edges is None:
        number_edges = SAMPLE_RANDOM_EDGE_COUNT

    if number_seed_edges is None:
        number_seed_edges = SAMPLE_RANDOM_EDGE_SEED_COUNT

    if seed is not None:
        random.seed(seed)

    # Check if graph will sample full graph, and just return it if it would
    if graph.number_of_edges() <= number_edges:
        logger.info("sampled full graph")
        return graph.copy()

    logger.debug(
        "getting random sub-graph with %d seed edges, %d final edges, and seed=%s",
        number_seed_edges,
        number_edges,
        seed,
    )

    # Get initial graph with `number_seed_edges` edges
    result = get_graph_with_random_edges(graph, number_seed_edges)

    number_edges_remaining = number_edges - result.number_of_edges()
    _helper(
        result,
        graph,
        number_edges_remaining,
        node_blacklist=set(),  # This is the set of nodes that should no longer be chosen to grow from
        invert_degrees=invert_degrees,
    )

    logger.debug("removing isolated nodes")
    remove_isolated_nodes(result)

    return result



# -*- coding: utf-8 -*-

"""Mutations that induce a sub-graph."""

from . import (
    annotations,
    citation,
    neighborhood,
    paths,
    random_subgraph,
    upstream,
    utils,
)
from .annotations import *
from .citation import *
from .neighborhood import *
from .paths import *
from .random_subgraph import *
from .upstream import *
from .utils import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""This module holds the Pipeline class."""

import json
import logging
import types
from functools import wraps
from typing import Any, Dict, Iterable, List, Optional, TextIO, Tuple, Union

from .decorators import get_transformation, in_place_map, mapped, universe_map
from .exc import MetaValueError, MissingPipelineFunctionError, MissingUniverseError
from ..operations import node_intersection, union

__all__ = [
    "Pipeline",
]

logger = logging.getLogger(__name__)

META_UNION = "union"
META_INTERSECTION = "intersection"


def _get_protocol_tuple(data: Dict[str, Any]) -> Tuple[str, List, Dict]:
    """Convert a dictionary to a tuple."""
    return data["function"], data.get("args", []), data.get("kwargs", {})


class Pipeline:
    """Build and runs analytical pipelines on BEL graphs.

    Example usage:

    >>> from pybel import BELGraph
    >>> from pybel.struct.pipeline import Pipeline
    >>> from pybel.struct.mutation import enrich_protein_and_rna_origins, prune_protein_rna_origins
    >>> graph = BELGraph()
    >>> example = Pipeline()
    >>> example.append(enrich_protein_and_rna_origins)
    >>> example.append(prune_protein_rna_origins)
    >>> result = example.run(graph)

    """

    def __init__(self, protocol: Optional[Iterable[Dict]] = None):
        """Initialize the pipeline with an optional pre-defined protocol.

        :param protocol: An iterable of dictionaries describing how to transform a network
        """
        self.universe = None
        self.protocol = protocol or []

    def __len__(self):
        return len(self.protocol)

    def __iter__(self):
        return iter(self.protocol)

    @staticmethod
    def from_functions(functions) -> "Pipeline":
        """Build a pipeline from a list of functions.

        :param functions: A list of functions or names of functions
        :type functions: iter[((pybel.BELGraph) -> pybel.BELGraph) or ((pybel.BELGraph) -> None) or str]

        Example with function:

        >>> from pybel.struct.pipeline import Pipeline
        >>> from pybel.struct.mutation import remove_associations
        >>> pipeline = Pipeline.from_functions([remove_associations])

        Equivalent example with function names:

        >>> from pybel.struct.pipeline import Pipeline
        >>> pipeline = Pipeline.from_functions(['remove_associations'])

        Lookup by name is possible for built in functions, and those that have been registered correctly using one of
        the four decorators:

        1. :func:`pybel.struct.pipeline.transformation`,
        2. :func:`pybel.struct.pipeline.in_place_transformation`,
        3. :func:`pybel.struct.pipeline.uni_transformation`,
        4. :func:`pybel.struct.pipeline.uni_in_place_transformation`,
        """
        result = Pipeline()

        for func in functions:
            result.append(func)

        return result

    def _get_function(self, name: str):
        """Wrap a function with the universe and in-place.

        :param name: The name of the function
        :rtype: types.FunctionType
        :raises MissingPipelineFunctionError: If the functions is not registered
        """
        f = mapped.get(name)

        if f is None:
            raise MissingPipelineFunctionError("{} is not registered as a pipeline function".format(name))

        if name in universe_map and name in in_place_map:
            return self._wrap_in_place(self._wrap_universe(f))

        if name in universe_map:
            return self._wrap_universe(f)

        if name in in_place_map:
            return self._wrap_in_place(f)

        return f

    def append(self, name, *args, **kwargs) -> "Pipeline":
        """Add a function (either as a reference, or by name) and arguments to the pipeline.

        :param name: The name of the function
        :type name: str or (pybel.BELGraph -> pybel.BELGraph)
        :param args: The positional arguments to call in the function
        :param kwargs: The keyword arguments to call in the function
        :return: This pipeline for fluid query building
        :raises MissingPipelineFunctionError: If the function is not registered
        """
        if isinstance(name, types.FunctionType):
            return self.append(name.__name__, *args, **kwargs)
        elif isinstance(name, str):
            get_transformation(name)
        else:
            raise TypeError("invalid function argument: {}".format(name))

        av = {
            "function": name,
        }

        if args:
            av["args"] = args

        if kwargs:
            av["kwargs"] = kwargs

        self.protocol.append(av)
        return self

    def extend(self, protocol: Union[Iterable[Dict], "Pipeline"]) -> "Pipeline":
        """Add another pipeline to the end of the current pipeline.

        :param protocol: An iterable of dictionaries (or another Pipeline)
        :return: This pipeline for fluid query building

        Example:
        >>> p1 = Pipeline.from_functions(['enrich_protein_and_rna_origins'])
        >>> p2 = Pipeline.from_functions(['remove_pathologies'])
        >>> p1.extend(p2)

        """
        for data in protocol:
            name, args, kwargs = _get_protocol_tuple(data)
            self.append(name, *args, **kwargs)

        return self

    def _run_helper(self, graph, protocol: Iterable[Dict]):
        """Help run the protocol.

        :param pybel.BELGraph graph: A BEL graph
        :param protocol: The protocol to run, as JSON
        :rtype: pybel.BELGraph
        """
        result = graph

        for entry in protocol:
            meta_entry = entry.get("meta")

            if meta_entry is None:
                name, args, kwargs = _get_protocol_tuple(entry)
                func = self._get_function(name)
                result = func(result, *args, **kwargs)
            else:
                networks = (self._run_helper(graph, subprotocol) for subprotocol in entry["pipelines"])

                if meta_entry == META_UNION:
                    result = union(networks)

                elif meta_entry == META_INTERSECTION:
                    result = node_intersection(networks)

                else:
                    raise MetaValueError("invalid meta-command: {}".format(meta_entry))

        return result

    def run(self, graph, universe=None):
        """Run the contained protocol on a seed graph.

        :param pybel.BELGraph graph: The seed BEL graph
        :param pybel.BELGraph universe: Allows just-in-time setting of the universe in case it wasn't set before.
                                        Defaults to the given network.
        :return: The new graph is returned if not applied in-place
        :rtype: pybel.BELGraph
        """
        self.universe = universe or graph.copy()
        return self._run_helper(graph.copy(), self.protocol)

    def __call__(self, graph, universe=None):
        """Call :meth:`Pipeline.run`.

        :param pybel.BELGraph graph: The seed BEL graph
        :param pybel.BELGraph universe: Allows just-in-time setting of the universe in case it wasn't set before.
                                        Defaults to the given network.
        :param bool in_place: Should the graph be copied before applying the algorithm?
        :return: The new graph is returned if not applied in-place
        :rtype: pybel.BELGraph

        Using __call__ allows for methods to be chained together then applied

        >>> from pybel.struct.mutation import remove_associations, remove_pathologies
        >>> from pybel.struct.pipeline.pipeline import Pipeline
        >>> from pybel import BELGraph
        >>> pipe = Pipeline.from_functions([remove_associations, remove_pathologies])
        >>> graph = BELGraph() ...
        >>> new_graph = pipe(graph)
        """
        return self.run(graph=graph, universe=universe)

    def _wrap_universe(self, func):  # noqa: D202
        """Take a function that needs a universe graph as the first argument and returns a wrapped one."""

        @wraps(func)
        def wrapper(graph, *args, **kwargs):
            """Apply the enclosed function with the universe given as the first argument."""
            if self.universe is None:
                raise MissingUniverseError(
                    "Can not run universe function [{}] - No universe is set".format(func.__name__),
                )

            return func(self.universe, graph, *args, **kwargs)

        return wrapper

    @staticmethod
    def _wrap_in_place(func):  # noqa: D202
        """Take a function that doesn't return the graph and returns the graph."""

        @wraps(func)
        def wrapper(graph, *args, **kwargs):
            """Apply the enclosed function and returns the graph."""
            func(graph, *args, **kwargs)
            return graph

        return wrapper

    def to_json(self) -> List:
        """Return this pipeline as a JSON list."""
        return self.protocol

    def dumps(self, **kwargs) -> str:
        """Dump this pipeline as a JSON string."""
        return json.dumps(self.to_json(), **kwargs)

    def dump(self, file: TextIO, **kwargs) -> None:
        """Dump this protocol to a file in JSON."""
        return json.dump(self.to_json(), file, **kwargs)

    @staticmethod
    def from_json(data: List) -> "Pipeline":
        """Build a pipeline from a JSON list."""
        return Pipeline(data)

    @staticmethod
    def load(file: TextIO) -> "Pipeline":
        """Load a protocol from JSON contained in file.

        :return: The pipeline represented by the JSON in the file
        :raises MissingPipelineFunctionError: If any functions are not registered
        """
        return Pipeline.from_json(json.load(file))

    @staticmethod
    def loads(s: str) -> "Pipeline":
        """Load a protocol from a JSON string.

        :param s: A JSON string
        :return: The pipeline represented by the JSON in the file
        :raises MissingPipelineFunctionError: If any functions are not registered
        """
        return Pipeline.from_json(json.loads(s))

    def __str__(self):
        return json.dumps(self.protocol, indent=2)

    @staticmethod
    def _build_meta(meta: str, pipelines: Iterable["Pipeline"]) -> "Pipeline":
        """Build a pipeline with a given meta-argument.

        :param meta: either union or intersection
        :param pipelines:
        """
        return Pipeline(
            protocol=[
                {
                    "meta": meta,
                    "pipelines": [pipeline.protocol for pipeline in pipelines],
                },
            ],
        )

    @staticmethod
    def union(pipelines: Iterable["Pipeline"]) -> "Pipeline":
        """Take the union of multiple pipelines.

        :param pipelines: A list of pipelines
        :return: The union of the results from multiple pipelines
        """
        return Pipeline._build_meta(META_UNION, pipelines)

    @staticmethod
    def intersection(pipelines: Iterable["Pipeline"]) -> "Pipeline":
        """Take the intersection of the results from multiple pipelines.

        :param pipelines: A list of pipelines
        :return: The intersection of results from multiple pipelines
        """
        return Pipeline._build_meta(META_INTERSECTION, pipelines)



# -*- coding: utf-8 -*-

"""Transport functions for `BioDati <https://biodati.com/>`_.

BioDati is a paid, closed-source platform for hosting BEL content. However,
they do have a demo instance running at https://studio.demo.biodati.com with which the examples
in this module will be described.

As noted in the transport functions for BioDati, you should change the URLs to point to your own
instance of BioDati. If you're looking for an open source storage system for  hosting your own BEL content,
you may consider `BEL Commons <https://github.com/bel-commons/bel-commons>`_, with the caveat that it is
currently maintained in an academic capacity. Disclosure: BEL Commons is developed by the developers of PyBEL.
"""

import json
import logging
from io import BytesIO
from typing import Any, Iterable, List, Mapping, Optional, Union

import requests
from more_itertools import chunked

from .graphdati import _iter_graphdati, from_graphdati, to_graphdati
from ..struct import BELGraph

__all__ = [
    "to_biodati",
    "from_biodati",
]

logger = logging.getLogger(__name__)


def to_biodati(  # noqa: S107
    graph: BELGraph,
    *,
    username: str = "demo@biodati.com",
    password: str = "demo",
    base_url: str = "https://nanopubstore.demo.biodati.com",
    chunksize: Optional[int] = None,
    use_tqdm: bool = True,
    collections: Optional[Iterable[str]] = None,
    overwrite: bool = False,
    validate: bool = True,
    email: Union[bool, str] = False,
) -> requests.Response:
    """Post this graph to a BioDati server.

    :param graph: A BEL graph
    :param username: The email address to log in to BioDati. Defaults to "demo@biodati.com" for the demo server
    :param password: The password to log in to BioDati. Defaults to "demo" for the demo server
    :param base_url: The BioDati nanopub store base url. Defaults to "https://nanopubstore.demo.biodati.com" for
     the demo server's nanopub store
    :param chunksize: The number of nanopubs to post at a time. By default, does all.
    :param use_tqdm: Should tqdm be used when iterating?
    :param collections: Tags to add to the nanopubs for lookup on BioDati
    :param overwrite: Set the BioDati upload "overwrite" setting
    :param validate: Set the BioDati upload "validate" setting
    :param email: Who should get emailed with results about the upload? If true, emails to user
     used for login. If string, emails to that user. If false, no email.
    :return: The response from the BioDati server (last response if using chunking)

    .. warning::

        BioDati does not support large uploads (yet?).

    .. warning::

        The default public BioDati server has been put here. You should
        switch it to yours. It will look like ``https://nanopubstore.<YOUR NAME>.biodati.com``.
    """
    biodati_client = BiodatiClient(username, password, base_url)
    if chunksize:
        return biodati_client.post_graph_chunked(
            graph,
            chunksize,
            use_tqdm=use_tqdm,
            collections=collections,
            overwrite=overwrite,
            validate=validate,
            email=email,
        )
    else:
        return biodati_client.post_graph(
            graph,
            use_tqdm=use_tqdm,
            collections=collections,
            overwrite=overwrite,
            validate=validate,
            email=email,
        )


def from_biodati(  # noqa: S107
    network_id: str,
    username: str = "demo@biodati.com",
    password: str = "demo",
    base_url: str = "https://networkstore.demo.biodati.com",
) -> BELGraph:
    """Get a graph from a BioDati network store based on its network identifier.

    :param network_id: The internal identifier of the network you want to download.
    :param username: The email address to log in to BioDati. Defaults to "demo@biodati.com" for the demo server
    :param password: The password to log in to BioDati. Defaults to "demo" for the demo server
    :param base_url: The BioDati network store base url. Defaults to "https://networkstore.demo.biodati.com" for
     the demo server's network store

    Example usage:

    .. code-block:: python

        from pybel import from_biodati
        network_id = '01E46GDFQAGK5W8EFS9S9WMH12'  # COVID-19 graph example from Wendy Zimmermann
        graph = from_biodati(
            network_id=network_id,
            username='demo@biodati.com',
            password='demo',
            base_url='https://networkstore.demo.biodati.com',
        )
        graph.summarize()

    .. warning::

        The default public BioDati server has been put here. You should
        switch it to yours. It will look like ``https://networkstore.<YOUR NAME>.biodati.com``.
    """
    biodati_client = BiodatiClient(username, password, base_url)
    return biodati_client.get_graph(network_id)


class BiodatiClient:
    """A client for the BioDati nanopub store and network store's APIs."""

    def __init__(self, username: str, password: str, base_url: str):
        self.base_url = base_url.rstrip("/")
        self.username = username
        res = requests.post(
            "{}/token".format(base_url),
            data=dict(username=username, password=password),
        )
        token_dict = res.json()
        self.token_type = token_dict["token_type"]
        self.id_token = token_dict["id_token"]
        self.access_token = token_dict["access_token"]

    def post(self, endpoint: str, **kwargs):
        """Send a post request to BioDati."""
        return self._help_request(requests.post, endpoint, **kwargs)

    def get(self, endpoint: str, **kwargs):
        """Send a get request to BioDati."""
        return self._help_request(requests.get, endpoint, **kwargs)

    def _help_request(self, requester, endpoint: str, **kwargs):
        """Send a request to BioDati."""
        url = "{}/{}".format(self.base_url, endpoint)
        logger.info("requesting %s with params %s", url, kwargs.get("params", {}))
        headers = {"Authorization": "{} {}".format(self.token_type, self.id_token)}
        return requester(url, headers=headers, **kwargs)

    def get_graph(self, network_id: str) -> BELGraph:
        """Get a graph from BioDati."""
        return from_graphdati(self.get_graph_json(network_id))

    def get_graph_json(self, network_id: str, network_format: str = "normal"):
        """Get the graph JSON."""
        res = self.get(
            "networks/{network_id}".format(network_id=network_id),
            params={"format": network_format},
        )
        # FIXME network_format='full' causes internal server error currently
        res_json = res.json()
        return res_json

    def post_graph(
        self,
        graph: BELGraph,
        *,
        use_tqdm: bool = True,
        collections: Optional[List[str]] = None,
        overwrite: bool = False,
        validate: bool = True,
        email: Union[bool, str] = False,
    ) -> requests.Response:
        """Post the graph to BioDati.

        :param graph: A BEL graph
        :param use_tqdm: Should tqdm be used when iterating?
        :param collections: Tags to add to the nanopubs for lookup on BioDati
        :param overwrite: Set the BioDati upload "overwrite" setting
        :param validate: Set the BioDati upload "validate" setting
        :param email: Who should get emailed with results about the upload? If true, emails to user
         used for login. If string, emails to that user. If false, no email.
        :return: Last response from upload
        """
        metadata_extras = dict()
        if collections is not None:
            metadata_extras.update(collections=list(collections))
        j = to_graphdati(graph, use_tqdm=use_tqdm, metadata_extras=metadata_extras)
        return self.post_graph_json(
            j,
            overwrite=overwrite,
            validate=validate,
            email=email,
        )

    def post_graph_chunked(
        self,
        graph: BELGraph,
        chunksize: int,
        *,
        use_tqdm: bool = True,
        collections: Optional[Iterable[str]] = None,
        overwrite: bool = False,
        validate: bool = True,
        email: Union[bool, str] = False,
    ) -> requests.Response:
        """Post the graph to BioDati in chunks, when the graph is too big for a normal upload.

        :param graph: A BEL graph
        :param chunksize: The size of the chunks of nanopubs to upload
        :param use_tqdm: Should tqdm be used when iterating?
        :param collections: Tags to add to the nanopubs for lookup on BioDati
        :param overwrite: Set the BioDati upload "overwrite" setting
        :param validate: Set the BioDati upload "validate" setting
        :param email: Who should get emailed with results about the upload? If true, emails to user
         used for login. If string, emails to that user. If false, no email.
        :return: Last response from upload
        """
        metadata_extras = dict()
        if collections is not None:
            metadata_extras.update(collections=list(collections))
        iterable = _iter_graphdati(graph, use_tqdm=use_tqdm, metadata_extras=metadata_extras)
        res = None
        for chunk in chunked(iterable, chunksize):
            res = self.post_graph_json(
                chunk,
                overwrite=overwrite,
                validate=validate,
                email=email,
            )
        return res

    def post_graph_json(
        self,
        graph_json: List[Mapping[str, Any]],
        overwrite: bool = False,
        validate: bool = True,
        email: Union[bool, str] = False,
    ) -> requests.Response:
        """Post the GraphDati object to BioDati.

        :param graph_json: The JSON object (in GraphDati schema) to upload to BioDati
        :param overwrite: Set the BioDati upload "overwrite" setting
        :param validate: Set the BioDati upload "validate" setting
        :param email: Who should get emailed with results about the upload? If true, emails to user
         used for login. If string, emails to that user. If false, no email.
        """
        file = BytesIO()
        file.write(json.dumps(graph_json).encode("utf-8"))
        file.seek(0)
        return self.post_graph_file(
            file,
            overwrite=overwrite,
            validate=validate,
            email=email,
        )

    def post_graph_file(
        self,
        file: BytesIO,
        overwrite: bool = False,
        validate: bool = True,
        email: Union[bool, str] = False,
    ) -> requests.Response:
        """Post a graph to BioDati.

        :param file: A file in bytes mode or BytesIO object
        :param overwrite: Set the BioDati upload "overwrite" setting
        :param validate: Set the BioDati upload "validate" setting
        :param email: Who should get emailed with results about the upload? If true, emails to user
         used for login. If string, emails to that user. If false, no email.
        """
        params = dict(overwrite=overwrite, validate=validate)
        if isinstance(email, str):
            params["email"] = email
        elif email:
            params["email"] = self.username

        return self.post(
            "nanopubs/import/file",
            files=dict(file=file),
            params=params,
        )


def _main():
    """Run with python -m pybel.io.graphdati."""
    network_id = "01E46GDFQAGK5W8EFS9S9WMH12"
    graph = from_biodati(network_id=network_id)
    graph.summarize()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    _main()



# -*- coding: utf-8 -*-

"""Language constants for BEL.

This module contains mappings between PyBEL's internal constants and BEL language keywords.
"""

import warnings
from typing import Optional

from .constants import (
    ABUNDANCE,
    BIOPROCESS,
    CELL_SURFACE,
    COMPLEX,
    COMPOSITE,
    EXTRACELLULAR,
    GENE,
    IDENTIFIER,
    INTRACELLULAR,
    MIRNA,
    NAME,
    NAMESPACE,
    PATHOLOGY,
    PROTEIN,
    RNA,
    TRANSCRIBED_TO,
    TRANSLATED_TO,
)
from .utils import ensure_quotes


class Entity(dict):
    """Represents a named entity with a namespace and name/identifier."""

    def __init__(
        self,
        *,
        namespace: str,
        name: Optional[str] = None,
        identifier: Optional[str] = None,
    ) -> None:
        """Create a dictionary representing a reference to an entity.

        :param namespace: The namespace to which the entity belongs
        :param name: The name of the entity
        :param identifier: The identifier of the entity in the namespace
        """
        if name is None and identifier is None:
            raise ValueError("cannot create an entity with neither a name nor identifier")
        if not isinstance(namespace, str):
            raise TypeError("namespace should be a string: {}".format(namespace))
        if not namespace:
            raise ValueError("namespace should be non-empty")

        super().__init__(
            {
                NAMESPACE: namespace,
            }
        )

        if name is not None:
            if not isinstance(name, str):
                raise TypeError("name should be a string: {}".format(name))
            if not name:
                raise ValueError("name should be non-empty")
            self[NAME] = name

        if identifier is not None:
            if not isinstance(identifier, str):
                raise TypeError(f"identifier should be a string. Got {type(identifier)} {identifier}")
            if not identifier:
                raise ValueError("identifier should be non-empty")
            self[IDENTIFIER] = identifier

    @property
    def namespace(self) -> str:  # noqa: D401
        """The entity's namespace."""
        return self[NAMESPACE]

    @property
    def name(self) -> str:  # noqa: D401
        """The entity's name or label."""
        return self.get(NAME)

    @property
    def identifier(self) -> str:  # noqa: D401
        """The entity's identifier."""
        return self.get(IDENTIFIER)

    @property
    def curie(self) -> str:
        """Return this entity as a CURIE."""
        return "{}:{}".format(
            self.namespace,
            ensure_quotes(self.identifier if self.identifier else self.name),
        )

    @property
    def obo(self) -> str:
        """Return this entity as an OBO-style CURIE."""
        return "{}:{} ! {}".format(
            self.namespace,
            ensure_quotes(self.identifier),
            ensure_quotes(self.name),
        )

    def __str__(self):  # noqa: D105
        return self.obo if self.identifier and self.name else self.curie

    def __hash__(self) -> int:
        return hash((self.namespace, self.identifier, self.name))


text_location_labels = {
    "Abstract": Entity(namespace="iao", identifier="0000315", name="abstract"),
    "Review": Entity(namespace="iao", identifier="0000311", name="publication"),  # sue me
    "Results": Entity(namespace="iao", identifier="0000318", name="results section"),
    "Legend": Entity(namespace="sio", identifier="000468 ", name="legend"),
}

#: A dictionary of activity labels used in the ma() function in activity(p(X), ma(Y))
activity_labels = {
    "catalyticActivity": "cat",
    "cat": "cat",
    "chaperoneActivity": "chap",
    "chap": "chap",
    "gtpBoundActivity": "gtp",
    "gtp": "gtp",
    "kinaseActivity": "kin",
    "kin": "kin",
    "peptidaseActivity": "pep",
    "pep": "pep",
    "phosphataseActivity": "phos",
    "phos": "phos",
    "ribosylationActivity": "ribo",
    "ribo": "ribo",
    "transcriptionalActivity": "tscript",
    "tscript": "tscript",
    "transportActivity": "tport",
    "tport": "tport",
    "molecularActivity": "molecularActivity",
    # Added by PyBEL
    "guanineNucleotideExchangeFactorActivity": "gef",
    "gef": "gef",
    "gtpaseActivatingProteinActivity": "gap",
    "gap": "gap",
}

#: Maps the default BEL molecular activities to Gene Ontology Molecular Functions
activity_mapping = {
    "act": Entity(namespace="go", name="molecular function", identifier="0003674"),
    "cat": Entity(namespace="go", name="catalytic activity", identifier="0003824"),
    "chap": Entity(
        namespace="go",
        name="protein binding involved in protein folding",
        identifier="0044183",
    ),
    "gtp": Entity(namespace="go", name="GTP binding", identifier="0005525"),
    "kin": Entity(namespace="go", name="kinase activity", identifier="0016301"),
    "pep": Entity(namespace="go", name="peptidase activity", identifier="0008233"),
    "phos": Entity(namespace="go", name="phosphatase activity", identifier="0016791"),
    "ribo": Entity(
        namespace="go",
        name="NAD(P)+-protein-arginine ADP-ribosyltransferase activity",
        identifier="0003956",
    ),
    "tscript": Entity(
        namespace="go",
        name="nucleic acid binding transcription factor activity",
        identifier="0001071",
    ),
    "tport": Entity(namespace="go", name="transporter activity", identifier="0005215"),
    "molecularActivity": Entity(namespace="go", name="molecular_function", identifier="0003674"),
    "gef": Entity(
        namespace="go",
        name="guanyl-nucleotide exchange factor activity",
        identifier="0005085",
    ),
    "gap": Entity(namespace="go", name="GTPase activating protein binding", identifier="0032794"),
}

activities = list(activity_labels.keys())

cytoplasm = Entity(name="cytoplasm", namespace="go", identifier="0005737")
nucleus = Entity(name="nucleus", namespace="go", identifier="0005634")
intracellular = Entity(name="intracellular", namespace="go", identifier="0005622")
extracellular = Entity(name="extracellular space", namespace="go", identifier="0005615")
cell_surface = Entity(name="cell surface", namespace="go", identifier="0009986")

#: Maps the default BEL cellular components to Gene Ontology Cellular Components
compartment_mapping = {
    INTRACELLULAR: intracellular,
    EXTRACELLULAR: extracellular,
    CELL_SURFACE: cell_surface,
    "cytoplasm": cytoplasm,
    "nucleus": nucleus,
}

compartments = list(compartment_mapping)

#: Provides a mapping from BEL terms to PyBEL internal constants
abundance_labels = {
    "abundance": ABUNDANCE,
    "a": ABUNDANCE,
    "geneAbundance": GENE,
    "g": GENE,
    "microRNAAbundance": MIRNA,
    "m": MIRNA,
    "proteinAbundance": PROTEIN,
    "p": PROTEIN,
    "rnaAbundance": RNA,
    "r": RNA,
    "biologicalProcess": BIOPROCESS,
    "bp": BIOPROCESS,
    "pathology": PATHOLOGY,
    "path": PATHOLOGY,
    "composite": COMPOSITE,
    "compositeAbundance": COMPOSITE,
    "complex": COMPLEX,
    "complexAbundance": COMPLEX,
}

#: Maps the BEL abundance types to the Systems Biology Ontology
abundance_sbo_mapping = {
    MIRNA: Entity(namespace="sbo", name="microRNA", identifier="0000316"),
    BIOPROCESS: Entity(namespace="sbo", name="process", identifier="0000375"),
    GENE: Entity(namespace="sbo", name="gene", identifier="0000243"),
    RNA: Entity(namespace="sbo", name="messenger RNA", identifier="0000278"),
    COMPLEX: Entity(namespace="sbo", name="protein complex", identifier="0000297"),
    PATHOLOGY: Entity(namespace="sbo", name="phenotype", identifier="0000358"),
}

relation_sbo_mapping = {
    TRANSLATED_TO: Entity(namespace="sbo", name="translation", identifier="0000184"),
    TRANSCRIBED_TO: Entity(namespace="sbo", name="transcription", identifier="0000183"),
}

amino_acid_dict = {
    "A": "Ala",
    "R": "Arg",
    "N": "Asn",
    "D": "Asp",
    "C": "Cys",
    "E": "Glu",
    "Q": "Gln",
    "G": "Gly",
    "H": "His",
    "I": "Ile",
    "L": "Leu",
    "K": "Lys",
    "M": "Met",
    "F": "Phe",
    "P": "Pro",
    "S": "Ser",
    "T": "Thr",
    "W": "Trp",
    "Y": "Tyr",
    "V": "Val",
}

dna_nucleotide_labels = {
    "A": "Adenine",
    "T": "Thymine",
    "C": "Cytosine",
    "G": "Guanine",
}

rna_nucleotide_labels = {
    "a": "adenine",
    "u": "uracil",
    "c": "cytosine",
    "g": "guanine",
}

#: A dictionary of default protein modifications to their preferred value
pmod_namespace = {
    "Ac": "Ac",
    "acetylation": "Ac",
    "ADPRib": "ADPRib",
    "ADP-ribosylation": "ADPRib",
    "adenosine diphosphoribosyl": "ADPRib",
    "Farn": "Farn",
    "farnesylation": "Farn",
    "Gerger": "Gerger",
    "geranylgeranylation": "Gerger",
    "Glyco": "Glyco",
    "glycosylation": "Glyco",
    "Hy": "Hy",
    "hydroxylation": "Hy",
    "ISG": "ISG",
    "ISGylation": "ISG",
    "ISG15-protein conjugation": "ISG",
    "Me": "Me",
    "methylation": "Me",
    "Me1": "Me1",
    "monomethylation": "Me1",
    "mono-methylation": "Me1",
    "Me2": "Me2",
    "dimethylation": "Me2",
    "di-methylation": "Me2",
    "Me3": "Me3",
    "trimethylation": "Me3",
    "tri-methylation": "Me3",
    "Myr": "Myr",
    "myristoylation": "Myr",
    "Nedd": "Nedd",
    "neddylation": "Nedd",
    "NGlyco": "NGlyco",
    "N-linked glycosylation": "NGlyco",
    "NO": "NO",
    "Nitrosylation": "NO",
    "OGlyco": "OGlyco",
    "O-linked glycosylation": "OGlyco",
    "Palm": "Palm",
    "palmitoylation": "Palm",
    "Ph": "Ph",
    "phosphorylation": "Ph",
    "Sulf": "Sulf",
    "sulfation": "Sulf",
    "sulphation": "Sulf",
    "sulfur addition": "Sulf",
    "sulphur addition": "Sulf",
    "sulfonation": "sulfonation",
    "sulphonation": "sulfonation",
    "Sumo": "Sumo",
    "SUMOylation": "Sumo",
    "Ub": "Ub",
    "ubiquitination": "Ub",
    "ubiquitinylation": "Ub",
    "ubiquitylation": "Ub",
    "UbK48": "UbK48",
    "Lysine 48-linked polyubiquitination": "UbK48",
    "UbK63": "UbK63",
    "Lysine 63-linked polyubiquitination": "UbK63",
    "UbMono": "UbMono",
    "monoubiquitination": "UbMono",
    "UbPoly": "UbPoly",
    "polyubiquitination": "UbPoly",
    # PyBEL Variants
    "Ox": "Ox",
    "oxidation": "Ox",
}

#: Use Gene Ontology children of go_0006464: "cellular protein modification process"
pmod_mappings = {
    "Ac": {
        "synonyms": ["Ac", "acetylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0006473", name="protein acetylation"),
            Entity(namespace="mod", identifier="00394", name="acetylated residue"),
            Entity(namespace="mop", identifier="0000030", name="acetylation"),
            Entity(namespace="sbo", identifier="0000215", name="acetylation"),
        ],
    },
    "ADPRib": {
        "synonyms": [
            "ADPRib",
            "ADP-ribosylation",
            "ADPRib",
            "ADP-rybosylation",
            "adenosine diphosphoribosyl",
        ],
        "xrefs": [
            Entity(namespace="go", identifier="0006471", name="protein ADP-ribosylation"),
            Entity(
                namespace="mod",
                identifier="00752",
                name="adenosine diphosphoribosyl (ADP-ribosyl) modified residue",
            ),
            Entity(
                namespace="mop",
                identifier="0000220",
                name="adenosinediphosphoribosylation",
            ),
        ],
    },
    "Farn": {
        "synonyms": ["Farn", "farnesylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018343", name="protein farnesylation"),
            Entity(namespace="mod", identifier="00437", name="farnesylated residue"),
            Entity(namespace="mop", identifier="0000429", name="farnesylation"),
        ],
    },
    "Gerger": {
        "synonyms": ["Gerger", "geranylgeranylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018344", name="protein geranylgeranylation"),
            Entity(namespace="mod", identifier="00441", name="geranylgeranylated residue "),
            Entity(namespace="mop", identifier="0000431", name="geranylgeranylation"),
        ],
    },
    "Glyco": {
        "synonyms": ["Glyco", "glycosylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0006486", name="protein glycosylation"),
            Entity(namespace="mod", identifier="00693", name="glycosylated residue"),
            Entity(namespace="mop", identifier="0000162", name="glycosylation"),
        ],
    },
    "Hy": {
        "synonyms": ["Hy" "hydroxylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018126", name="protein hydroxylation"),
            Entity(namespace="mod", identifier="00677", name="hydroxylated residue"),
            Entity(namespace="mop", identifier="0000673", name="hydroxylation"),
        ],
    },
    "ISG": {
        "synonyms": ["ISG", "ISGylation", "ISG15-protein conjugation"],
        "xrefs": [
            Entity(namespace="go", identifier="0032020", name="ISG15-protein conjugation"),
        ],
        "activities": [
            Entity(namespace="go", identifier="0042296", name="ISG15 transferase activity"),
        ],
    },
    "Me": {
        "synonyms": ["Me", "methylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0006479", name="protein methylation"),
            Entity(namespace="mod", identifier="00427", name="methylated residue"),
        ],
    },
    "Me1": {
        "synonyms": ["Me1", "monomethylation", "mono-methylation"],
        "xrefs": [
            Entity(namespace="mod", identifier="00599", name="monomethylated residue"),
        ],
        "is_a": ["Me"],
    },
    "Me2": {
        "synonyms": ["Me2", "dimethylation", "di-methylation"],
        "xrefs": [
            Entity(namespace="mod", identifier="00429", name="dimethylated residue"),
        ],
        "is_a": ["Me"],
    },
    "Me3": {
        "synonyms": ["Me3", "trimethylation", "tri-methylation"],
        "xrefs": [
            Entity(namespace="mod", identifier="00430", name="trimethylated residue"),
        ],
        "is_a": ["Me"],
    },
    "Myr": {
        "synonyms": ["Myr", "myristoylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018377", name="protein myristoylation"),
            Entity(namespace="mod", identifier="00438", name="myristoylated residue"),
        ],
    },
    "Nedd": {
        "synonyms": ["Nedd", "neddylation", "RUB1-protein conjugation"],
        "xrefs": [
            Entity(namespace="go", identifier="0045116", name="protein neddylation"),
            Entity(namespace="mod", identifier="01150", name="neddylated lysine"),
        ],
    },
    "NGlyco": {
        "synonyms": ["NGlyco", "N-linked glycosylation"],
        "xrefs": [
            Entity(
                namespace="go",
                identifier="0006487",
                name="protein N-linked glycosylation",
            ),
            Entity(namespace="mod", identifier="00006", name="N-glycosylated residue"),
            Entity(namespace="mop", identifier="0002162", name="N-glycosylation"),
        ],
        "is_a": ["Glyco"],
    },
    "NO": {
        "synonyms": ["NO", "Nitrosylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0017014", name="protein nitrosylation"),
        ],
    },
    "Ox": {
        "synonyms": ["Ox", "oxidation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018158", name="protein oxidation"),
        ],
    },
    "OGlyco": {
        "synonyms": ["OGlyco", "O-linked glycosylation"],
        "xrefs": [
            Entity(
                namespace="go",
                identifier="0006493",
                name="protein O-linked glycosylation",
            ),
            Entity(namespace="mod", identifier="00396", name="O-glycosylated residue"),
            Entity(namespace="mop", identifier="0003162", name="O-glycosylation"),
        ],
        "is_a": ["Glyco"],
    },
    "Palm": {
        "synonyms": ["Palm", "palmitoylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0018345", name="protein palmitoylation"),
            Entity(namespace="mod", identifier="00440", name="palmitoylated residue"),
        ],
    },
    "Ph": {
        "synonyms": ["Ph", "phosphorylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0006468", name="protein phosphorylation"),
            Entity(namespace="mod", identifier="00696"),
        ],
    },
    "Sulf": {
        "synonyms": [
            "Sulf",
            "sulfation",
            "sulphation",
            "sulfur addition",
            "sulphur addition",
            "sulfonation",
            "sulphonation",
        ],
        "xrefs": [
            Entity(namespace="go", identifier="0006477", name="protein sulfation"),
            Entity(namespace="mod", identifier="00695", name="sulfated residue"),
            Entity(namespace="mop", identifier="0000559", name="sulfonation"),
        ],
        "target": [
            Entity(namespace="chebi", identifier="29922", name="sulfo group"),
        ],
    },
    "Sumo": {
        "synonyms": ["Sumo", "SUMOylation", "Sumoylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0016925", name="protein sumoylation"),
            Entity(namespace="mod", identifier="01149", name="sumoylated lysine"),
        ],
        "activities": [
            Entity(namespace="go", identifier="0019789", name="SUMO transferase activity"),
        ],
    },
    "Ub": {
        "synonyms": ["Ub", "ubiquitination", "ubiquitinylation", "ubiquitylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0016567", name="protein ubiquitination"),
            Entity(namespace="mod", identifier="01148", name="ubiquitinylated lysine"),
            Entity(namespace="sbo", identifier="0000224", name="ubiquitination"),
        ],
    },
    "UbK48": {
        "synonyms": ["UbK48", "Lysine 48-linked polyubiquitination"],
        "xrefs": [
            Entity(
                namespace="go",
                identifier="0070936",
                name="protein K48-linked ubiquitination",
            ),
        ],
    },
    "UbK63": {
        "synonyms": ["UbK63", "Lysine 63-linked polyubiquitination"],
        "xrefs": [
            Entity(
                namespace="go",
                identifier="0070534",
                name="protein K63-linked ubiquitination",
            ),
        ],
    },
    "UbMono": {
        "synonyms": ["UbMono", "monoubiquitination"],
        "xrefs": [
            Entity(namespace="go", identifier="0006513", name="protein monoubiquitination"),
        ],
    },
    "UbPoly": {
        "synonyms": ["UbPoly", "polyubiquitination"],
        "xrefs": [
            Entity(namespace="go", identifier="0000209", name="protein polyubiquitination"),
        ],
    },
}

#: A dictionary of legacy (BEL 1.0) default namespace protein modifications to their BEL 2.0 preferred value
pmod_legacy_labels = {
    "P": "Ph",
    "A": "Ac",
    "F": "Farn",
    "G": "Glyco",
    "H": "Hy",
    "M": "Me",
    "R": "ADPRib",
    "S": "Sumo",
    "U": "Ub",
    "O": "Ox",
}

#: A dictionary of default gene modifications. This is a PyBEL variant to the BEL specification.
gmod_namespace = {
    "methylation": "Me",
    "Me": "Me",
    "M": "Me",
    "ADPRib": "ADPRib",
}

#: Use Gene Ontology children of go:0006304 ! "DNA modification"
gmod_mappings = {
    "Me": {
        "synonyms": ["Me", "M", "methylation"],
        "xrefs": [
            Entity(namespace="go", identifier="0006306", name="DNA methylation"),
        ],
    },
    "ADPRib": {
        "synonyms": ["ADPRib"],
        "xrefs": [
            Entity(namespace="go", identifier="0030592", name="DNA ADP-ribosylation"),
        ],
    },
}


class CitationDict(Entity):
    """A dictionary describing a citation."""

    def __init__(self, namespace: str, identifier: str, *, name: Optional[str] = None, **kwargs):
        super().__init__(namespace=namespace, identifier=identifier, name=name)
        self.update(kwargs)


def citation_dict(
    *,
    namespace: Optional[str] = None,
    db: Optional[str] = None,
    identifier: Optional[str] = None,
    db_id: Optional[str] = None,
    name: Optional[str] = None,
    **kwargs,
) -> CitationDict:
    """Make a citation dictionary."""
    if namespace and db:
        raise ValueError("can not specify both namespace and db")
    if identifier and db_id:
        raise ValueError("can not specify both identifier and db_id")
    if db:
        warnings.warn(
            "usage of keyword argument `db` in citation_dict() should be replaced with `namespace`. "
            "Will be removed in PyBEL 16.",
            DeprecationWarning,
        )
        namespace = db
    if db_id:
        warnings.warn(
            "usage of keyword argument `db_id` in citation_dict() should be replaced with `identifier`. "
            "Will be removed in PyBEL 16.",
            DeprecationWarning,
        )
        identifier = db_id

    return CitationDict(namespace=namespace, identifier=identifier, name=name, **kwargs)



# -*- coding: utf-8 -*-

"""Modules supporting deletion and degradation of graphs."""

from . import deletion, protein_rna_origins
from .deletion import *
from .protein_rna_origins import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""TSV conversion."""

import json
import logging
from typing import List, Optional, TextIO, Tuple, Union

from networkx.utils import open_file
from tqdm.autonotebook import tqdm

from . import converters
from ...dsl import BaseEntity
from ...struct import BELGraph

__all__ = [
    "to_triples_file",
    "to_edgelist",
    "to_triples",
    "to_triple",
]

logger = logging.getLogger(__name__)


class NoTriplesValueError(ValueError):
    """Raised when no triples could be converted."""


@open_file(1, mode="w")
def to_triples_file(
    graph: BELGraph, path: Union[str, TextIO], *, use_tqdm: bool = False, sep="\t", raise_on_none: bool = False
) -> None:
    """Write the graph as a TSV.

    :param graph: A BEL graph
    :param path: A path or file-like
    :param use_tqdm: Should a progress bar be shown?
    :param sep: The separator to use
    :param raise_on_none: Should an exception be raised if no triples are returned?
    :raises: NoTriplesValueError
    """
    for h, r, t in to_triples(graph, use_tqdm=use_tqdm, raise_on_none=raise_on_none):
        print(h, r, t, sep=sep, file=path)


@open_file(1, mode="w")
def to_edgelist(
    graph: BELGraph, path: Union[str, TextIO], *, use_tqdm: bool = False, sep="\t", raise_on_none: bool = False
) -> None:
    """Write the graph as an edgelist.

    :param graph: A BEL graph
    :param path: A path or file-like
    :param use_tqdm: Should a progress bar be shown?
    :param sep: The separator to use
    :param raise_on_none: Should an exception be raised if no triples are returned?
    :raises: NoTriplesValueError
    """
    for h, r, t in to_triples(graph, use_tqdm=use_tqdm, raise_on_none=raise_on_none):
        print(h, t, json.dumps(dict(relation=r)), sep=sep, file=path)


def to_triples(graph: BELGraph, use_tqdm: bool = False, raise_on_none: bool = False) -> List[Tuple[str, str, str]]:
    """Get a non-redundant list of triples representing the graph.

    :param graph: A BEL graph
    :param use_tqdm: Should a progress bar be shown?
    :param raise_on_none: Should an exception be raised if no triples are returned?
    :raises: NoTriplesValueError
    """
    it = graph.edges(keys=True)

    if use_tqdm:
        it = tqdm(
            it,
            total=graph.number_of_edges(),
            desc="Preparing TSV for {}".format(graph),
            unit_scale=True,
            unit="edge",
        )

    triples = (to_triple(graph, u, v, key) for u, v, key in it)

    # clean duplicates and Nones
    rv = list(
        sorted({triple for triple in triples if triple is not None}),
    )
    if raise_on_none and not rv:
        raise NoTriplesValueError("Could not convert any triples")
    return rv


def to_triple(
    graph: BELGraph,
    u: BaseEntity,
    v: BaseEntity,
    key: str,
) -> Optional[Tuple[str, str, str]]:  # noqa: C901
    """Get the triples' strings that should be written to the file."""
    data = graph[u][v][key]

    # order is important
    _converters = [
        converters.ListComplexHasComponentConverter,
        converters.PartOfNamedComplexConverter,
        converters.SubprocessPartOfBiologicalProcessConverter,
        converters.ProteinPartOfBiologicalProcessConverter,
        converters.AbundancePartOfPopulationConverter,
        converters.PopulationPartOfAbundanceConverter,
        converters.RegulatesActivityConverter,
        converters.MiRNADecreasesExpressionConverter,
        converters.MiRNADirectlyDecreasesExpressionConverter,
        converters.AbundanceDirectlyDecreasesProteinActivityConverter,
        converters.AbundanceDirectlyIncreasesProteinActivityConverter,
        converters.IsAConverter,
        converters.EquivalenceConverter,
        converters.CorrelationConverter,
        converters.AssociationConverter,
        converters.DrugIndicationConverter,
        converters.DrugSideEffectConverter,
        converters.RegulatesAmountConverter,
        converters.ProcessCausalConverter,
        converters.IncreasesAmountConverter,
        converters.DecreasesAmountConverter,
        converters.NoChangeAmountConverter,
        converters.IncreasesActivityConverter,
        converters.DecreasesActivityConverter,
        converters.NoChangeActivityConverter,
        converters.ReactionHasProductConverter,
        converters.ReactionHasReactantConverter,
        converters.ReactionHasCatalystConverter,
        converters.HasVariantConverter,
        converters.IncreasesDegradationConverter,
        converters.DecreasesDegradationConverter,
        converters.RegulatesDegradationConverter,
        converters.NoChangeDegradationConverter,
        converters.TranscriptionFactorForConverter,
        converters.HomomultimerConverter,
        converters.BindsGeneConverter,
        converters.BindsProteinConverter,
        converters.ProteinRegulatesComplex,
    ]

    for converter in _converters:
        if converter.predicate(u, v, key, data):
            return converter.convert(u, v, key, data)

    logger.warning("unhandled: {}".format(graph.edge_to_bel(u, v, data)))



# -*- coding: utf-8 -*-

import json

from pybel import BELGraph
from pybel.constants import INCREASES, RELATION
from pybel.dsl import hgnc
from pybel.manager.models import Edge, Namespace, NamespaceEntry, Network, Node
from pybel.testing.cases import TemporaryCacheMixin
from pybel.testing.mocks import mock_bel_resources
from pybel.testing.utils import make_dummy_annotations, make_dummy_namespaces, n
from tests.constants import test_citation_dict, test_evidence_text

yfg1 = hgnc(identifier="1", name="YFG1")
yfg2 = hgnc(identifier="2", name="YFG1")
yfg3 = hgnc(identifier="3", name="YFG3")


def make_increase_edge(u, v):
    bel = "{} {} {}".format(u.as_bel(), INCREASES, v.as_bel())
    data = json.dumps({RELATION: INCREASES})
    assert data
    return Edge(source=u, target=v, relation=INCREASES, bel=bel, data=data)


class TestReconstituteNodeTuples(TemporaryCacheMixin):
    @mock_bel_resources
    def test_simple(self, mock):
        """This test checks that the network can be added and dropped"""
        graph = BELGraph(name="test", version="0.0.0")
        graph.annotation_pattern["Disease"] = ".*"
        graph.annotation_pattern["Cell"] = ".*"
        graph.add_increases(
            yfg1,
            yfg2,
            evidence=test_evidence_text,
            citation=test_citation_dict,
            annotations={
                "Disease": {"Disease1": True},
                "Cell": {"Cell1": True},
            },
        )

        make_dummy_namespaces(self.manager, graph)
        make_dummy_annotations(self.manager, graph)

        network = self.manager.insert_graph(graph)

        self.manager.drop_network_by_id(network.id)


class TestCascades(TemporaryCacheMixin):
    def setUp(self):
        super(TestCascades, self).setUp()

        self.n1 = Node._start_from_base_entity(yfg1)
        self.n2 = Node._start_from_base_entity(yfg2)
        self.n3 = Node._start_from_base_entity(yfg3)
        self.e1 = make_increase_edge(self.n1, self.n2)
        self.e2 = make_increase_edge(self.n2, self.n3)
        self.e3 = make_increase_edge(self.n1, self.n3)
        self.g1 = Network(name=n(), version=n(), edges=[self.e1, self.e2, self.e3])
        self.g2 = Network(name=n(), version=n(), edges=[self.e1])

        self.manager.session.add_all([self.n1, self.n2, self.n3, self.e1, self.e2, self.e3, self.g1, self.g2])
        self.manager.session.commit()

        self.assertEqual(3, self.manager.count_nodes())
        self.assertEqual(3, self.manager.count_edges())
        self.assertEqual(2, self.manager.count_networks())

    def test_drop_node(self):
        """Makes sure that when a node gets dropped, its in-edges AND out-edges also do"""
        self.manager.session.delete(self.n2)
        self.manager.session.commit()

        self.assertEqual(2, self.manager.count_nodes())
        self.assertEqual(1, self.manager.count_edges())
        self.assertEqual(2, self.manager.count_networks())
        self.assertEqual(1, self.g1.edges.count())
        self.assertEqual(0, self.g2.edges.count())

    def test_drop_edge(self):
        """When an edge gets dropped, make sure the network doesn't have as many edges, but nodes get to stay"""
        self.manager.session.delete(self.e1)
        self.manager.session.commit()

        self.assertEqual(3, self.manager.count_nodes())
        self.assertEqual(2, self.manager.count_edges())
        self.assertEqual(2, self.manager.count_networks())
        self.assertEqual(2, self.g1.edges.count())
        self.assertEqual(0, self.g2.edges.count())

    def test_get_orphan_edges(self):
        edges = [result.edge_id for result in self.manager.query_singleton_edges_from_network(self.g1)]
        self.assertEqual(2, len(edges))
        self.assertIn(self.e2.id, edges)
        self.assertIn(self.e3.id, edges)

    def test_drop_network_1(self):
        """When a network gets dropped, drop all of the edges if they don't appear in other networks"""
        self.manager.drop_network(self.g1)

        self.assertEqual(3, self.manager.count_nodes())
        self.assertEqual(1, self.manager.count_edges())
        self.assertEqual(1, self.manager.count_networks())
        self.assertEqual(1, self.g2.edges.count())

    def test_drop_network_2(self):
        """When a network gets dropped, drop all of the edges if they don't appear in other networks"""
        self.manager.drop_network(self.g2)

        self.assertEqual(3, self.manager.count_nodes())
        self.assertEqual(3, self.manager.count_edges())
        self.assertEqual(1, self.manager.count_networks())
        self.assertEqual(3, self.g1.edges.count())

    def test_drop_all_networks(self):
        """When all networks are dropped, make sure all the edges and network_edge mappings are gone too"""
        self.manager.drop_networks()

        self.assertEqual(0, self.manager.count_edges())
        self.assertEqual(0, self.manager.count_networks())

    def test_drop_modification(self):
        """Don't let this happen"""

    def test_drop_property(self):
        """Don't let this happen"""

    def test_drop_namespace(self):
        keyword, url = n(), n()

        namespace = Namespace(keyword=keyword, url=url)
        self.manager.session.add(namespace)

        n_entries = 5

        for _ in range(n_entries):
            entry = NamespaceEntry(name=n(), namespace=namespace)
            self.manager.session.add(entry)

        self.manager.session.commit()

        self.assertEqual(1, self.manager.count_namespaces(), msg="Should have one namespace")
        self.assertEqual(
            n_entries,
            self.manager.count_namespace_entries(),
            msg="Should have {} entries".format(n_entries),
        )

        self.manager.drop_namespace_by_url(url)

        self.assertEqual(0, self.manager.count_namespaces(), msg="Should have no namespaces")
        self.assertEqual(
            0,
            self.manager.count_namespace_entries(),
            msg="Entries should have been dropped",
        )



# -*- coding: utf-8 -*-

"""Parsers for modifications to abundances."""

from .fragment import get_fragment_language
from .fusion import get_fusion_language, get_legacy_fusion_langauge
from .gene_modification import get_gene_modification_language
from .gene_substitution import get_gene_substitution_language
from .location import get_location_language
from .protein_modification import get_protein_modification_language
from .protein_substitution import get_protein_substitution_language
from .truncation import get_truncation_language
from .variant import get_hgvs_language



# -*- coding: utf-8 -*-

"""Resources for PyBEL."""

CHEBI_URL = "https://raw.githubusercontent.com/pharmacome/conso/d67144bc27a21626a514837b3b4382413dd6866b/external/chebi-names.belns"
EC_URL = (
    "https://raw.githubusercontent.com/pharmacome/conso/d67144bc27a21626a514837b3b4382413dd6866b/external/ec-code.belns"
)
FB_URL = "https://raw.githubusercontent.com/pharmacome/conso/80171ae62cf43aa1fc8a6c326b94537ab342458c/external/fb-names.belns"
GO_URL = "https://raw.githubusercontent.com/pharmacome/conso/d9d270e11aac480542c412d4222983a5f042b8ae/external/go-names.belns"
HGNC_URL = "https://raw.githubusercontent.com/pharmacome/conso/d67144bc27a21626a514837b3b4382413dd6866b/external/hgnc-names.belns"
MESH_URL = "https://raw.githubusercontent.com/pharmacome/conso/f02c6ad4a4791a8ed45448513b9de8c8f1b00c87/external/mesh-names.belns"
MGI_URL = "https://raw.githubusercontent.com/pharmacome/conso/efc856fb009a39e4d284269a6801f79ed3d3cf56/external/mgi-names.belns"
NCBIGENE_URL = "https://raw.githubusercontent.com/pharmacome/conso/d67144bc27a21626a514837b3b4382413dd6866b/external/ncbigene-names.belns"
RGD_URL = "https://raw.githubusercontent.com/pharmacome/conso/efc856fb009a39e4d284269a6801f79ed3d3cf56/external/rgd-names.belns"

FPLX_URL = (
    "https://raw.githubusercontent.com/sorgerlab/famplex/da9f2187b694e6b425e668604e24ac9fac0f2c31/export/famplex.belns"
)

#: Default URL lookup for some keywords
keyword_to_url = dict(
    chebi=CHEBI_URL,
    ec=EC_URL,
    fb=FB_URL,
    go=GO_URL,
    hgnc=HGNC_URL,
    mesh=MESH_URL,
    mgi=MGI_URL,
    ncbigene=NCBIGENE_URL,
    rgd=RGD_URL,
    fplx=FPLX_URL,
)



# -*- coding: utf-8 -*-

"""Exceptions for the :mod:`pybel.struct.pipeline` module."""

__all__ = [
    "MissingPipelineFunctionError",
    "MetaValueError",
    "MissingUniverseError",
    "DeprecationMappingError",
    "PipelineNameError",
]


class MissingPipelineFunctionError(KeyError):
    """Raised when trying to run the pipeline with a function that isn't registered."""


class MetaValueError(ValueError):
    """Raised when getting an invalid meta value."""


class MissingUniverseError(ValueError):
    """Raised when running a universe function without a universe being present."""


class DeprecationMappingError(ValueError):
    """Raised when applying the deprecation function annotation and the given name already is being used."""


class PipelineNameError(ValueError):
    """Raised when a second function tries to use the same name."""



# -*- coding: utf-8 -*-

"""Predicates for checking nodes' variants."""

from functools import wraps
from typing import Tuple, Type, Union

from .utils import node_predicate
from ..typing import NodePredicate
from ....dsl import (
    BaseEntity,
    CentralDogma,
    Fragment,
    GeneModification,
    Hgvs,
    ProteinModification,
    Variant,
)

__all__ = [
    "has_variant",
    "has_protein_modification",
    "has_gene_modification",
    "has_fragment",
    "has_hgvs",
]


@node_predicate
def has_variant(node: BaseEntity) -> bool:
    """Return true if the node has any variants."""
    return isinstance(node, CentralDogma) and node.variants


def _variant_checker(variant_cls: Union[Type[Variant], Tuple[Type[Variant], ...]]) -> NodePredicate:
    @node_predicate
    @wraps(node_has_variant)
    def _rv(node: BaseEntity):
        return node_has_variant(node, variant_cls)

    return _rv


def node_has_variant(node: BaseEntity, variant_cls) -> bool:
    """Return true if the node has at least one of the given variant."""
    return (
        isinstance(node, CentralDogma)
        and node.variants
        and any(isinstance(variant, variant_cls) for variant in node.variants)
    )


has_protein_modification = _variant_checker(ProteinModification)
has_gene_modification = _variant_checker(GeneModification)
has_hgvs = _variant_checker(Hgvs)
has_fragment = _variant_checker(Fragment)



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with a SQL database."""

import logging
from typing import Optional

from sqlalchemy.exc import IntegrityError, OperationalError

from .cache_manager import Manager

__all__ = [
    "to_database",
    "from_database",
]

logger = logging.getLogger(__name__)


def to_database(
    graph,
    manager: Optional[Manager] = None,
    use_tqdm: bool = True,
):
    """Store a graph in a database.

    :param BELGraph graph: A BEL graph
    :return: If successful, returns the network object from the database.
    :rtype: Optional[Network]
    """
    if manager is None:
        manager = Manager()

    try:
        return manager.insert_graph(graph, use_tqdm=use_tqdm)
    except (IntegrityError, OperationalError):
        manager.session.rollback()
        logger.exception("Error storing graph")
    except Exception as e:
        manager.session.rollback()
        raise e


def from_database(
    name: str,
    version: Optional[str] = None,
    manager: Optional[Manager] = None,
):
    """Load a BEL graph from a database.

    If name and version are given, finds it exactly with
    :meth:`pybel.manager.Manager.get_network_by_name_version`. If just the name is given, finds most recent
    with :meth:`pybel.manager.Manager.get_network_by_name_version`

    :param name: The name of the graph
    :param version: The version string of the graph. If not specified, loads most recent graph added with this name
    :return: A BEL graph loaded from the database
    :rtype: Optional[BELGraph]
    """
    if manager is None:
        manager = Manager()

    if version is None:
        return manager.get_graph_by_most_recent(name)

    return manager.get_graph_by_name_version(name, version)
